import pathlib
import re
import time
from typing import Any, Dict, List, Optional, Sequence, Tuple, TypeVar, Union

import numpy as np
import SimpleITK as sitk

from imgtools.crawler import crawl
from imgtools.io.loaders import (
    BaseLoader,
    ImageCSVLoader,
    ImageFileLoader,
    read_dicom_auto,
    read_image,
)
from imgtools.io.writers import (
    BaseSubjectWriter,
    BaseWriter,
    HDF5Writer,
    ImageFileWriter,
    MetadataWriter,
    NumpyWriter,
    SegNrrdWriter,
)
from imgtools.logging import logger
from imgtools.modules.datagraph import DataGraph
from imgtools.modules import Segmentation, map_over_labels, StructureSet
from imgtools.ops.functional import (
    bounding_box,
    centroid,
    clip_intensity,
    crop,
    crop_to_mask_bounding_box,
    image_statistics,
    min_max_scale,
    resample,
    resize,
    rotate,
    standard_scale,
    window_intensity,
    zoom,
)
from imgtools.utils import (
    array_to_image,
    image_to_array,
    physical_points_to_idxs,
)

LoaderFunction = TypeVar("LoaderFunction")
ImageFilter = TypeVar("ImageFilter")
Function = TypeVar("Function")


# Base class
class BaseOp:
    def __call__(self, *args, **kwargs):
        raise NotImplementedError

    def __repr__(self):
        attrs = [(k, v) for k, v in self.__dict__.items() if not k.startswith("_")]
        attrs = [(k, f"'{v}'") if isinstance(v, str) else (k, v) for k, v in attrs]
        args = ", ".join(f"{k}={v}" for k, v in attrs)
        return f"{self.__class__.__name__}({args})"


# Input/output
class BaseInput(BaseOp):
    def __init__(self, loader):
        if not isinstance(loader, BaseLoader):
            raise ValueError(
                f"loader must be a subclass of io.BaseLoader, got {type(loader)}"
            )
        self._loader = loader

    def __call__(self, key):
        inputs = self._loader.get(key)
        return inputs


class BaseOutput(BaseOp):
    def __init__(self, writer):
        if not isinstance(writer, BaseWriter):
            raise ValueError(
                f"writer must be a subclass of io.BaseWriter, got {type(writer)}"
            )
        self._writer = writer

    def __call__(self, key, *args, **kwargs):
        self._writer.put(key, *args, **kwargs)


class ImageAutoInput(BaseInput):
    """ImageAutoInput class is a wrapper class around ImgCSVloader which looks for the specified
    directory and crawls through it as the first step. Using the crawled output data, a graph on
    modalties present in the dataset is formed which stores the relation between all the modalities.
    Based on the user provided modalities, this class loads the information of the user provided modalities

    Parameters
    ----------
    dir_path: str
        Path to dataset top-level directory. The crawler/indexer will start at this directory.

    modalities: str | List[str]
        Either as a list of strings or a single string of comma-separated modalities.
        i.e ["CT", "RTSTRUCT"] or "CT,RTSTRUCT".
        Only samples with ALL modalities will be processed.
        Make sure there are no space between list elements as it is parsed as a string.

    n_jobs: int
        Number of parallel jobs to run when crawling. Default is -1.

    visualize: bool
        Whether to return visualization of the data graph. Requires pyvis to be installed.
        Default is False.

    update: bool
        Whether to update crawled index
    """

    imgtools_dir: str = ".imgtools"

    def __init__(
        self,
        dir_path: str,
        modalities: str | List[str],
        n_jobs: int = -1,
        visualize: bool = False,
        update: bool = False,
    ):
        self.modalities = modalities

        self.dir_path = pathlib.Path(dir_path)
        self.parent = self.dir_path.parent
        self.dataset_name = self.dir_path.name

        self.csv_path = (
            self.parent / self.imgtools_dir / f"imgtools_{self.dataset_name}.csv"
        )
        self.json_path = (
            self.parent / self.imgtools_dir / f"imgtools_{self.dataset_name}.json"
        )
        self.edge_path = (
            self.parent / self.imgtools_dir / f"imgtools_{self.dataset_name}_edges.csv"
        )

        start = time.time()
        # CRAWLER
        # -------
        # Checks if dataset has already been indexed
        if not self.csv_path.exists() or update:
            logger.debug("Output exists, force updating.", path=self.csv_path)
            logger.info("Indexing the dataset...")
            db = crawl(self.dir_path, n_jobs=n_jobs)
            logger.info(f"Number of patients in the dataset: {len(db)}")
        else:
            logger.warning(
                "The dataset has already been indexed. Use --update to force update."
            )

        # GRAPH
        # -----
        # Form the graph
        logger.debug("Creating edge path", edge_path=self.edge_path)
        self.graph = DataGraph(
            path_crawl=self.csv_path.resolve().as_posix(),
            edge_path=self.edge_path.as_posix(),
            visualize=visualize,
            update=update,
        )
        logger.info(
            f"Forming the graph based on the given modalities: {self.modalities}"
        )
        self.df_combined = self.graph.parser(self.modalities)

        self.output_streams = [
            ("_").join(cols.split("_")[1:])
            for cols in self.df_combined.columns
            if cols.split("_")[0] == "folder"
        ]

        # not sure what this is really doing...

        self.column_names = [
            cols for cols in self.df_combined.columns if cols.split("_")[0] == "folder"
        ]
        self.series_names = [
            cols for cols in self.df_combined.columns if cols.split("_")[0] == "series"
        ]
        logger.info(
            f"There are {len(self.df_combined)} cases containing all {self.modalities} modalities."
        )

        self.readers = [read_dicom_auto for _ in range(len(self.output_streams))]
        logger.info(f"Total time taken: {time.time() - start:.2f} seconds")
        loader = ImageCSVLoader(
            self.df_combined,
            colnames=self.column_names,
            seriesnames=self.series_names,
            id_column=None,
            expand_paths=False,
            readers=self.readers,
        )

        super().__init__(loader)


class ImageCSVInput(BaseInput):
    """ImageCSVInput class looks for a CSV file in a specified directory and loads images based on the information in the file.

    Parameters
    ----------
    csv_path: str
        Directory where the CSV file is stored.

    colnames: List[str]
        List of column names in the CSV file to be used for image loading.

    id_column: str, optional
        A column name to be used as the subject ID.

    reader: LoaderFunction
        The functions used to read images.
        The functions are implemented in the imgtools.io module.
        The options are:
        - read_image
        - read_dicom_series
        - read_dicom_rtstruct
        - read_segmentation
    """

    def __init__(
        self,
        csv_path_or_dataframe: str,
        colnames: List[str] = None,
        id_column: Optional[str] = None,
        expand_paths: bool = True,
        readers: List[LoaderFunction] = None,
    ):  # no mutable defaults: https://florimond.dev/en/posts/2018/08/python-mutable-defaults-are-the-source-of-all-evil/
        if colnames is None:
            colnames = []
        if readers is None:
            readers = [read_image]
        self.csv_path_or_dataframe = csv_path_or_dataframe
        self.colnames = colnames
        self.id_column = id_column
        self.expand_paths = expand_paths
        self.readers = readers
        loader = ImageCSVLoader(
            self.csv_path_or_dataframe,
            colnames=self.colnames,
            id_column=self.id_column,
            expand_paths=self.expand_paths,
            readers=self.readers,
        )
        super().__init__(loader)


class ImageFileInput(BaseInput):
    """ImageFileInput class looks for images in a specified directory and reads them by using a reader function.

    Parameters
    ----------
    root_directory: str
        Root directory where the image files are stored.

    get_subject_id_from: str
        How to extract the subject ID. The options are:
        - 'filename' indicates the image reader to use filename as the subject ID. (default)
        - 'subject_directory' indicates the image reader to use the name of the subject directory.

    subdir_path: str, optional
        Path to a subdirectory where the image is stored.

    exclude_paths: List[str], optional
        Directory paths to be excluded when searching for the image files to be loaded.

    reader: LoaderFunction
        The function used to read individual images.
        The functions are implemented in the imgtools.io module.
        The options are:
        - read_image
        - read_dicom_series
        - read_dicom_rtstruct
        - read_segmentation
        - read_dicom_auto
        - read_dicom_rtdose
        - read_dicom_pet
    """

    def __init__(
        self,
        root_directory: str,
        get_subject_id_from: str = "filename",
        subdir_path: Optional[str] = None,
        exclude_paths: Optional[
            List[str]
        ] = None,  # no mutable defaults https://florimond.dev/en/posts/2018/08/python-mutable-defaults-are-the-source-of-all-evil/
        reader: LoaderFunction = None,
    ):
        if exclude_paths is None:
            exclude_paths = []
        if reader is None:
            reader = read_image
        self.root_directory = root_directory
        self.get_subject_id_from = get_subject_id_from
        self.subdir_path = subdir_path
        self.exclude_paths = exclude_paths
        self.reader = reader
        loader = ImageFileLoader(
            self.root_directory,
            self.get_subject_id_from,
            self.subdir_path,
            self.exclude_paths,
            self.reader,
        )
        super().__init__(loader)


class ImageFileOutput(BaseOutput):
    """ImageFileOutput class outputs processed images as one of the image file formats.

    Parameters
    ----------
    root_directory: str
        Root directory where the processed image files will be stored.

    filename_format: str, optional
        The filename template.
        Set to be {subject_id}.nrrd as default.
        {subject_id} will be replaced by each subject's ID at runtime.

    create_dirs: bool, optional
        Specify whether to create an output directory if it does not exit.
        Set to be True as default.

    compress: bool, optional
        Specify whether to enable compression for NRRD format.
        Set to be true as default.
    """

    def __init__(
        self,
        root_directory: str,
        filename_format: Optional[str] = "{subject_id}.nrrd",
        create_dirs: Optional[bool] = True,
        compress: Optional[bool] = True,
    ):
        self.root_directory = root_directory
        self.filename_format = filename_format
        self.create_dirs = create_dirs
        self.compress = compress

        if ".seg" in filename_format:  # from .seg.nrrd bc it is now .nii.gz
            writer_class = SegNrrdWriter
        else:
            writer_class = ImageFileWriter

        writer = writer_class(
            self.root_directory, self.filename_format, self.create_dirs, self.compress
        )

        super().__init__(writer)


# Resampling ops
class ImageSubjectFileOutput(BaseOutput):
    def __init__(
        self,
        root_directory: str,
        filename_format: Optional[str] = "{subject_id}.nii.gz",
        create_dirs: Optional[bool] = True,
        compress: Optional[bool] = True,
    ):
        self.root_directory = root_directory
        self.filename_format = filename_format
        self.create_dirs = create_dirs
        self.compress = compress

        writer = BaseSubjectWriter(
            self.root_directory, self.filename_format, self.create_dirs, self.compress
        )

        super().__init__(writer)


class ImageAutoOutput:
    """
    Wrapper class around ImageFileOutput. This class supports multiple modalities writers and calls ImageFileOutput for writing the files

    Parameters
    ----------
    root_directory: str
        The directory where all the processed files will be stored in the form of nrrd

    output_streams: List[str]
        The modalties that should be stored. This is typically equal to the column names of the table returned after graph querying. Examples is provided in the
        dictionary file_name
    """

    def __init__(
        self,
        root_directory: str,
        output_streams: List[str],
        nnunet_info: Dict = None,
        inference: bool = False,
    ):
        self.output = {}
        for colname in output_streams:
            # Not considering colnames ending with alphanumeric
            colname_process = ("_").join(
                [item for item in colname.split("_") if not item.isnumeric()]
            )
            colname_process = colname  # temproary force #
            if not nnunet_info and not inference:
                self.output[colname_process] = ImageSubjectFileOutput(
                    pathlib.Path(
                        root_directory, "{subject_id}", colname_process.split(".")[0]
                    ).as_posix(),
                    filename_format="{}.nii.gz".format(colname_process),
                )
            elif inference:
                self.output[colname_process] = ImageSubjectFileOutput(
                    root_directory,
                    filename_format="{subject_id}_{modality_index}.nii.gz",
                )
            else:
                self.output[colname_process] = ImageSubjectFileOutput(
                    pathlib.Path(
                        root_directory, "{label_or_image}{train_or_test}"
                    ).as_posix(),
                    filename_format="{subject_id}_{modality_index}.nii.gz",
                )

    def __call__(
        self,
        subject_id: str,
        img: sitk.Image,
        output_stream,
        is_mask: bool = False,
        mask_label: Optional[str] = "",
        label_or_image: str = "images",
        train_or_test: str = "Tr",
        nnunet_info: Dict = None,
    ):
        self.output[output_stream](
            subject_id,
            img,
            is_mask=is_mask,
            mask_label=mask_label,
            label_or_image=label_or_image,
            train_or_test=train_or_test,
            nnunet_info=nnunet_info,
        )


class NumpyOutput(BaseOutput):
    """NumpyOutput class processed images as NumPy files.

    Parameters
    ----------
    root_directory: str
        Root directory where the processed NumPy files will be stored.

    filename_format: str, optional
        The filename template.
        Set to be {subject_id}.npy as default.
        {subject_id} will be replaced by each subject's ID at runtime.

    create_dirs: bool, optional
        Specify whether to create an output directory if it does not exit.
        Set to be True as default.

    """

    def __init__(
        self,
        root_directory: str,
        filename_format: Optional[str] = "{subject_id}.npy",
        create_dirs: Optional[bool] = True,
    ):
        self.root_directory = root_directory
        self.filename_format = filename_format
        self.create_dirs = create_dirs
        writer = NumpyWriter(
            self.root_directory, self.filename_format, self.create_dirs
        )
        super().__init__(writer)


class HDF5Output(BaseOutput):
    """HDF5Output class outputs the processed image data in HDF5 format.

    Parameters
    ----------
    root_directory: str
        Root directory where the processed .h5 file will be stored.

    filename_format: str, optional
        The filename template.
        Set to be {subject_id}.h5 as default.
        {subject_id} will be replaced by each subject's ID at runtime.

    create_dirs: bool, optional
        Specify whether to create an output directory if it does not exit.
        Set to be True as default.

    save_geometry: bool, optional
        Specify whether to save geometry data.
        Set to be True as default.

    """

    def __init__(
        self,
        root_directory: str,
        filename_format: Optional[str] = "{subject_id}.h5",
        create_dirs: Optional[bool] = True,
        save_geometry: Optional[bool] = True,
    ):
        self.root_directory = root_directory
        self.filename_format = filename_format
        self.create_dirs = create_dirs
        self.save_geometry = save_geometry
        writer = HDF5Writer(
            self.root_directory,
            self.filename_format,
            self.create_dirs,
            self.save_geometry,
        )
        super().__init__(writer)


class MetadataOutput(BaseOutput):
    """MetadataOutput class outputs the metadata of processed image files in .json format.

    Parameters
    ----------
    root_directory: str
        Root directory where the processed .json file will be stored.

    filename_format: str, optional
        The filename template.
        Set to be {subject_id}.json as default.
        {subject_id} will be replaced by each subject's ID at runtime.

    create_dirs: bool, optional
        Specify whether to create an output directory if it does not exit.
        Set to be True as default.

    """

    def __init__(
        self,
        root_directory: str,
        filename_format: Optional[str] = "{subject_id}.json",
        create_dirs: Optional[bool] = True,
    ):
        self.root_directory = root_directory
        self.filename_format = filename_format
        self.create_dirs = create_dirs
        writer = MetadataWriter(
            self.root_directory, self.filename_format, self.create_dirs
        )
        super().__init__(writer)


# Resampling ops
class Resample(BaseOp):
    """Resample operation class:
    A callable class that resamples image to a given spacing, optionally applying a transformation.

    To instantiate:
        obj = Resample(spacing, interpolation, anti_alias, anti_alias_sigma, transform, output_size)

    To call:
        result = obj(image)

    Parameters
    ----------
    spacing
        The new image spacing. If float, assumes the same spacing in all
        directions. Alternatively, a sequence of floats can be passed to
        specify spacing along each dimension. Passing 0 at any position will
        keep the original spacing along that dimension (useful for in-plane
        resampling).

    interpolation, optional
        The interpolation method to use. Valid options are:
        - "linear" for bi/trilinear interpolation (default)
        - "nearest" for nearest neighbour interpolation
        - "bspline" for order-3 b-spline interpolation

    anti_alias, optional
        Whether to smooth the image with a Gaussian kernel before resampling.
        Only used when downsampling, i.e. when `spacing < image.GetSpacing()`.
        This should be used to avoid aliasing artifacts.

    anti_alias_sigma, optional
        The standard deviation of the Gaussian kernel used for anti-aliasing.

    transform, optional
        Transform to apply to input coordinates when resampling. If None,
        defaults to identity transformation.

    output_size, optional
        Size of the output image. If None, it is computed to preserve the
        whole extent of the input image.
    """

    def __init__(
        self,
        spacing: Union[float, Sequence[float], np.ndarray],
        interpolation: str = "linear",
        anti_alias: bool = True,
        anti_alias_sigma: Optional[float] = None,
        transform: Optional[sitk.Transform] = None,
        output_size: Optional[Sequence[float]] = None,
    ):
        self.spacing = spacing
        self.interpolation = interpolation
        self.anti_alias = anti_alias
        self.anti_alias_sigma = anti_alias_sigma
        self.transform = transform
        self.output_size = output_size

    def __call__(self, image: sitk.Image) -> sitk.Image:
        """Resample callable object:
        Resamples image to a given spacing, optionally applying a transformation..

        Parameters
        ----------
        image
            The image to resample.

        Returns
        -------
        sitk.Image
            The resampled image.
        """

        return resample(
            image,
            spacing=self.spacing,
            interpolation=self.interpolation,
            anti_alias=self.anti_alias,
            anti_alias_sigma=self.anti_alias_sigma,
            transform=self.transform,
            output_size=self.output_size,
        )


class Resize(BaseOp):
    """Resize operation class:
    A callable class that resizes image to a given size by resampling coordinates.

    To instantiate:
        obj = Resize(size, interpolation, anti_alias, anti_alias_sigma)

    To call:
        result = obj(image)

    Parameters
    ----------
    size
        The new image size. If float, assumes the same size in all directions.
        Alternatively, a sequence of floats can be passed to specify size along
        each dimension. Passing 0 at any position will keep the original
        size along that dimension.

    interpolation, optional
        The interpolation method to use. Valid options are:
        - "linear" for bi/trilinear interpolation (default)
        - "nearest" for nearest neighbour interpolation
        - "bspline" for order-3 b-spline interpolation

    anti_alias, optional
        Whether to smooth the image with a Gaussian kernel before resampling.
        Only used when downsampling, i.e. when `size < image.GetSize()`.
        This should be used to avoid aliasing artifacts.

    anti_alias_sigma, optional
        The standard deviation of the Gaussian kernel used for anti-aliasing.
    """

    def __init__(
        self,
        size: Union[int, Sequence[int], np.ndarray],
        interpolation: str = "linear",
        anti_alias: bool = True,
        anti_alias_sigma: Optional[float] = None,
    ):
        self.size = size
        self.interpolation = interpolation
        self.anti_alias = anti_alias
        self.anti_alias_sigma = anti_alias_sigma

    def __call__(self, image: sitk.Image) -> sitk.Image:
        """Resize callable object: Resizes image to a given size by resampling coordinates.

        Parameters
        ----------
        image
            The image to resize.

        Returns
        -------
        sitk.Image
            The resized image.
        """

        return resize(
            image,
            size=self.size,
            interpolation=self.interpolation,
            anti_alias_sigma=self.anti_alias_sigma,
        )


class Zoom(BaseOp):
    """Zoom operation class: A callable class that rescales image, preserving its spatial extent.

    To instantiate:
        obj = Zoom(scale_factor, interpolation, anti_alias, anti_alias_sigma)

    To call:
        result = obj(image)

    The rescaled image will have the same spatial extent (size) but will be
    rescaled by `scale_factor` in each dimension. Alternatively, a separate
    scale factor for each dimension can be specified by passing a sequence
    of floats.

    Parameters
    ----------
    scale_factor
        If float, each dimension will be scaled by that factor. If tuple, each
        dimension will be scaled by the corresponding element.

    interpolation, optional
        The interpolation method to use. Valid options are:
        - "linear" for bi/trilinear interpolation (default)
        - "nearest" for nearest neighbour interpolation
        - "bspline" for order-3 b-spline interpolation

    anti_alias, optional
        Whether to smooth the image with a Gaussian kernel before resampling.
        Only used when downsampling, i.e. when `size < image.GetSize()`.
        This should be used to avoid aliasing artifacts.

    anti_alias_sigma, optional
        The standard deviation of the Gaussian kernel used for anti-aliasing.
    """

    def __init__(
        self,
        scale_factor: Union[float, Sequence[float]],
        interpolation: str = "linear",
        anti_alias: bool = True,
        anti_alias_sigma: Optional[float] = None,
    ):
        self.scale_factor = scale_factor
        self.interpolation = interpolation
        self.anti_alias = anti_alias
        self.anti_alias_sigma = anti_alias_sigma

    def __call__(self, image: sitk.Image) -> sitk.Image:
        """Zoom callable object: Rescales image, preserving its spatial extent.

        The rescaled image will have the same spatial extent (size) but will be
        rescaled by `scale_factor` in each dimension. Alternatively, a separate
        scale factor for each dimension can be specified by passing a sequence
        of floats.

        Parameters
        ----------
        image
            The image to rescale.

        Returns
        -------
        sitk.Image
            The rescaled image.
        """

        return zoom(
            image,
            self.scale_factor,
            interpolation=self.interpolation,
            anti_alias=self.anti_alias,
            anti_alias_sigma=self.anti_alias_sigma,
        )


class Rotate(BaseOp):
    """Rotate operation class: A callable class that rotates an image around a given centre.

    To instantiate:
        obj = Rotate(rotation_centre, angles, interpolation)

    To call:
        result = obj(image)

    Parameters
    ----------
    rotation_centre
        The centre of rotation in image coordinates.

    angles
        The angles of rotation around x, y and z axes.

    interpolation, optional
        The interpolation method to use. Valid options are:
        - "linear" for bi/trilinear interpolation (default)
        - "nearest" for nearest neighbour interpolation
        - "bspline" for order-3 b-spline interpolation
    """

    def __init__(
        self,
        rotation_centre: Sequence[float],
        angles: Union[float, Sequence[float]],
        interpolation: str = "linear",
    ):
        self.rotation_centre = rotation_centre
        self.angles = angles
        self.interpolation = interpolation

    def __call__(self, image: sitk.Image) -> sitk.Image:
        """Rotate callable object: Rotates an image around a given centre.

        Parameters
        ----------
        image
            The image to rotate.

        Returns
        -------
        sitk.Image
            The rotated image.
        """

        return rotate(
            image,
            rotation_centre=self.rotation_centre,
            angles=self.angles,
            interpolation=self.interpolation,
        )


class InPlaneRotate(BaseOp):
    """InPlaneRotate operation class: A callable class that rotates an image on a plane.

    To instantiate:
        obj = InPlaneRotate(angle, interpolation)

    To call:
        result = obj(image)

    Parameters
    ----------
    angle
        The angle of rotation.

    interpolation, optional
        The interpolation method to use. Valid options are:
        - "linear" for bi/trilinear interpolation (default)
        - "nearest" for nearest neighbour interpolation
        - "bspline" for order-3 b-spline interpolation
    """

    def __init__(self, angle: float, interpolation: str = "linear"):
        self.angle = angle
        self.interpolation = interpolation

    def __call__(self, image: sitk.Image) -> sitk.Image:
        """InPlaneRotate callable object: Rotates an image on a plane.

        Parameters
        ----------
        image
            The image to rotate.

        Returns
        -------
        sitk.Image
            The rotated image.
        """

        image_size = np.array(image.GetSize())
        image_centre = image_size // 2
        angles = (0.0, 0.0, self.angle)
        return rotate(
            image,
            rotation_centre=image_centre.tolist(),
            angles=angles,
            interpolation=self.interpolation,
        )


# Cropping & mask ops
class Crop(BaseOp):
    """Crop operation class: A callable class that crops an image to
    the desired size around a given centre.

    To instantiate:
        obj = Crop(crop_centre, size)

    To call:
        result = obj(image)

    Note that the cropped image might be smaller than size in a particular
    direction if the cropping window exceeds image boundaries.

    Parameters
    ----------

    crop_centre
        The centre of the cropping window in image coordinates.

    size
        The size of the cropping window along each dimension in pixels. If
        float, assumes the same size in all directions. Alternatively, a
        sequence of floats can be passed to specify size along x, y and z
        dimensions. Passing 0 at any position will keep the original size along
        that dimension.
    """

    def __init__(
        self, crop_centre: Sequence[float], size: Union[int, Sequence[int], np.ndarray]
    ):
        self.crop_centre = crop_centre
        self.size = size

    def __call__(self, image) -> sitk.Image:
        """Crop callable object: Crops an image to the desired size around a given centre.

        Note that the cropped image might be smaller than size in a particular
        direction if the cropping window exceeds image boundaries.

        Parameters
        ----------
        image
            The image to crop.

        Returns
        -------
        sitk.Image
            The cropped image.
        """

        return crop(image, crop_centre=self.crop_centre, size=self.size)


class CentreCrop(BaseOp):
    """CentreCrop operation class: A callable class that crops an image to the desired size
    around the centre of an image.

    To instantiate:
        obj = CentreCrop(size)

    To call:
        result = obj(image)

    Note that the cropped image might be smaller than size in a particular
    direction if the cropping window exceeds image boundaries.

    Parameters
    ----------
    size
        The size of the cropping window along each dimension in pixels. If
        float, assumes the same size in all directions. Alternatively, a
        sequence of floats can be passed to specify size along x, y and z
        dimensions. Passing 0 at any position will keep the original size along
        that dimension.
    """

    def __init__(self, size: Union[int, Sequence[int]]):
        self.size = size

    def __call__(self, image: sitk.Image) -> sitk.Image:
        """CentreCrop callable object: Crops an image to the desired size
        around the centre of an image.

        Note that the cropped image might be smaller than size in a particular
        direction if the cropping window exceeds image boundaries.

        Parameters
        ----------
        image
            The image to crop.

        Returns
        -------
        sitk.Image
            The cropped image.
        """
        image_size = np.array(image.GetSize())
        image_centre = image_size // 2
        return crop(image, crop_centre=image_centre, size=self.size)


class BoundingBox(BaseOp):
    """BoundingBox opetation class: A callable class that find the axis-aligned
    bounding box of a region descriibed by a segmentation mask.

    To instantiate:
        obj = BoundingBox()

    To call:
        result = obj(mask, label)
    """

    def __call__(self, mask: sitk.Image, label: int = 1) -> Tuple[Tuple, Tuple]:
        """BoundingBox callable object: Find the axis-aligned
        bounding box of a region descriibed by a segmentation mask.

        Parameters
        ----------
        mask
            Segmentation mask describing the region of interest. Can be an image of
            type unsigned int representing a label map or `segmentation.Segmentation`.

        label, optional
            Label to use when computing bounding box if segmentation mask contains
            more than 1 labelled region.

        Returns
        -------
        tuple of tuples
            The bounding box location and size. The first tuple gives the
            coordinates of the corner closest to the origin and the second
            gives the size in pixels along each dimension.
        """

        return bounding_box(mask, label=label)


class Centroid(BaseOp):
    """Centroid operation class: A callable class that finds the centroid of
    a labelled region specified by a segmentation mask.

    To instantiate:
        obj = Centroid(world_coordinates)

    To call:
        result = obj(mask, label)

    Parameters
    ----------
    world_coordinates, optional
        If True, return centroid in world coordinates, otherwise in image
        (voxel) coordinates (default).
    """

    def __init__(self, world_coordinates: bool = False):
        self.world_coordinates = world_coordinates

    def __call__(self, mask: sitk.Image, label: Optional[int] = 1) -> tuple:
        """Centroid callable object: Finds the centroid of
        a labelled region specified by a segmentation mask.

        Parameters
        ----------
        mask
            Segmentation mask describing the region of interest. Can be an image of
            type unsigned int representing a label map or `segmentation.Segmentation`.

        label, optional
            Label to use when computing the centroid if segmentation mask contains
            more than 1 labelled region.

        Returns
        -------
        tuple
            The centroid coordinates.
        """

        return centroid(mask, label=label, world_coordinates=self.world_coordinates)


class CropToMaskBoundingBox(BaseOp):
    """CropToMaskBoundingBox opetation class:
    A callable class that crops the image using the bounding box of a region of interest specified
    by a segmentation mask.

    To instantiate:
        obj = CropToMaskBoundingBox(margin)

    To call:
        result = obj(image, mask, label)

    Parameters
    ----------
    margin
        A margin that will be added to each dimension when cropping. If int,
        add the same margin to each dimension. A sequence of ints can also be
        passed to specify the margin separately along each dimension.
    """

    def __init__(self, margin: Union[int, Sequence[int], np.ndarray]):
        self.margin = margin

    def __call__(
        self,
        image: sitk.Image,
        mask: Union[int, Sequence[int], np.ndarray] = None,
        label: Optional[int] = 1,
    ) -> Tuple[sitk.Image]:
        """CropToMaskBoundingBox callable object:
        Crops the image using the bounding box of a region of interest specified
        by a segmentation mask.

        Parameters
        ----------
        image
            The image to crop.

        mask
            Segmentation mask describing the region of interest. Can be an image of
            type unsigned int representing a label map or `segmentation.Segmentation`.

        label, optional
            Label to use when computing the centroid if segmentation mask contains
            more than 1 labelled region.

        Returns
        -------
        tuple of sitk.Image
            The cropped image and mask.
        """

        return crop_to_mask_bounding_box(image, mask, margin=self.margin, label=label)


# Intensity ops
class ClipIntensity(BaseOp):
    """ClipIntensity operation class:
    A callable class that clips image grey level intensities to specified range.

    To instantiate:
        obj = ClipIntensity(lower, upper)

    To call:
        result = obj(image)

    The grey level intensities in the resulting image will fall in the range
    [lower, upper].

    Parameters
    ----------
    lower
        The lower bound on grey level intensity. Voxels with lower intensity
        will be set to this value.

    upper
        The upper bound on grey level intensity. Voxels with higer intensity
        will be set to this value.
    """

    def __init__(self, lower: float, upper: float):
        self.lower = lower
        self.upper = upper

    def __call__(self, image: sitk.Image) -> sitk.Image:
        """ClipIntensity callable object:
        Clips image grey level intensities to specified range.

        Parameters
        ----------
        image
            The intensity image to clip.

        Returns
        -------
        sitk.Image
            The clipped intensity image.
        """
        return clip_intensity(image, self.lower, self.upper)


class WindowIntensity(BaseOp):
    """WindowIntensity operation class:
    A callable class that restricts image grey level intensities to a given window and level.

    To instantiate:
        obj = WindowIntensity(window, level)

    To call:
        result = obj(image)

    The grey level intensities in the resulting image will fall in the range
    [level - window / 2, level + window / 2].

    Parameters
    ----------
    window
        The width of the intensity window.

    level
        The mid-point of the intensity window.
    """

    def __init__(self, window: float, level: float):
        self.window = window
        self.level = level

    def __call__(self, image: sitk.Image) -> sitk.Image:
        """WindowIntensity callable object:
        Restricts image grey level intensities to a given window and level.

        Parameters
        ----------
        image
            The intensity image to window.

        Returns
        -------
        sitk.Image
            The windowed intensity image.
        """

        return window_intensity(image, self.window, self.level)


class ImageStatistics(BaseOp):
    """ImageStatistics operation class:
    A callable class that computes the intensity statistics of an image.

    To instantiate:
        obj = ImageStatistics()

    To call:
        result = obj(image, mask, label)

    Returns the minimum, maximum, sum, mean, variance and standard deviation
    of image intensities.
    This function also supports computing the statistics in a specific
    region of interest if `mask` and `label` are passed.
    """

    def __call__(
        self,
        image: sitk.Image,
        mask: Optional[sitk.Image] = None,
        label: Optional[int] = 1,
    ) -> float:
        """ImageStatistics callable object:
        Computes the intensity statistics of an image.

        Returns the minimum, maximum, sum, mean, variance and standard deviation
        of image intensities.
        This function also supports computing the statistics in a specific
        region of interest if `mask` and `label` are passed.

        Parameters
        ----------
        image
            The image used to compute the statistics.

        mask, optional
            Segmentation mask specifying a region of interest used in computation.
            Can be an image of type unsigned int representing a label map or
            `segmentation.Segmentation`. Only voxels falling within the ROI will
            be considered. If None, use the whole image.

        label, optional
            Label to use when computing the statistics if segmentation mask contains
            more than 1 labelled region.

        Returns
        -------
        collections.namedtuple
            The computed intensity statistics in the image or region.
        """

        return image_statistics(image, mask, label=label)


class StandardScale(BaseOp):
    """StandardScale operation class:
    A callable class that rescales image intensities by subtracting
    the mean and dividing by standard deviation.

    To instantiate:
        obj = StandardScale(rescale_mean, rescale_std)
    To call
        result = obj(image, mask, label)

    If `rescale_mean` and `rescale_std` are None, image mean and standard
    deviation will be used, i.e. the resulting image intensities will have
    0 mean and unit variance. Alternatively, a specific mean and standard
    deviation can be passed to e.g. standardize a whole dataset of images.
    If a segmentation mask is passed, only the voxels falling within the mask
    will be considered when computing the statistics. However, the whole image
    will still be normalized using the computed values.

    Parameters
    ----------

    rescale_mean, optional
        The mean intensity used in rescaling. If None, image mean will be used.

    rescale_std, optional
        The standard deviation used in rescaling. If None, image standard
        deviation will be used.
    """

    def __init__(
        self, rescale_mean: Optional[float] = 0.0, rescale_std: Optional[float] = 1.0
    ):
        self.rescale_mean = rescale_mean
        self.rescale_std = rescale_std

    def __call__(
        self,
        image: sitk.Image,
        mask: Optional[sitk.Image] = None,
        label: Optional[int] = 1,
    ) -> sitk.Image:
        """StandardScale callable object:
        A callable class that rescales image intensities by subtracting
        the mean and dividing by standard deviation.

        Parameters
        ----------
        image
            sitk.Image object to be rescaled.

        mask, optional
            Segmentation mask specifying a region of interest used in computation.
            Can be an image of type unsigned int representing a label map or
            `segmentation.Segmentation`. Only voxels falling within the ROI will
            be considered. If None, use the whole image.

        label, optional
            Label to use when computing the statistics if segmentation mask contains
            more than 1 labelled region.

        Returns
        -------
        sitk.Image
            The rescaled image.
        """

        return standard_scale(image, mask, self.rescale_mean, self.rescale_std, label)


class MinMaxScale(BaseOp):
    """MinMaxScale operation class:
    A callable class that rescales image intensities to a given minimum and maximum.

    Applies a linear transformation to image intensities such that the minimum
    and maximum intensity values in the resulting image are equal to minimum
    (default 0) and maximum (default 1) respectively.

    To instantiante:
        obj = MinMaxScale(minimum, maximum)
    To call:
        result = obj(image)

    Parameters
    ----------

    minimum, optional
        The minimum intensity in the rescaled image.

    maximum, optional
        The maximum intensity in the rescaled image.
    """

    def __init__(self, minimum: float, maximum: float):
        self.minimum = minimum
        self.maximum = maximum

    def __call__(self, image: sitk.Image) -> sitk.Image:
        """MinMaxScale callable object:
        Rescales image intensities to a given minimum and maximum.

        Applies a linear transformation to image intensities such that the minimum
        and maximum intensity values in the resulting image are equal to minimum
        (default 0) and maximum (default 1) respectively.

        Parameters
        ----------
        image
            sitk.Image object to be rescaled.

        Returns
        -------
        sitk.Image
            The rescaled image.
        """
        return min_max_scale(image, self.minimum, self.maximum)


# Lambda ops


class SimpleITKFilter(BaseOp):
    """SimpleITKFilter operation class:
    A callable class that accepts an sitk.ImageFilter object to add a filter to an image.

    To instantiate:
        obj = SimpleITKFilter(sitk_filter, *execute_args)
    To call:
        result = obj(image)

    Parameters
    ----------
    sitk_filter
        An ImageFilter object in sitk library.

    execute_args, optional
        Any arguments to be passed to the Execute() function of the selected ImageFilter object.
    """

    def __init__(self, sitk_filter: ImageFilter, *execute_args: Optional[Any]):
        self.sitk_filter = sitk_filter
        self.execute_args = execute_args

    def __call__(self, image: sitk.Image) -> sitk.Image:
        """SimpleITKFilter callable object:
        A callable class that uses an sitk.ImageFilter object to add a filter to an image.

        Parameters
        ----------
        image
            sitk.Image object to be processed.

        Returns
        -------
        sitk.Image
            The processed image with a given filter.
        """
        return self.sitk_filter.Execute(image, *self.execute_args)


class ImageFunction(BaseOp):
    """ImageFunction operation class:
    A callable class that takens in a function to be used to process an image,
    and executes it.

    To instantiate:
        obj = ImageFunction(function, copy_geometry, **kwargs)
    To call:
        result = obj(image)

    Parameters
    ----------
    function
        A function to be used for image processing.
        This function needs to have the following signature:
        - function(image: sitk.Image, **args)
        - The first argument needs to be an sitkImage, followed by optional arguments.

    copy_geometry, optional
        An optional argument to specify whether information about the image should be copied to the
        resulting image. Set to be true as a default.

    kwargs, optional
        Any number of arguements used in the given function.
    """

    def __init__(
        self, function: Function, copy_geometry: bool = True, **kwargs: Optional[Any]
    ):
        self.function = function
        self.copy_geometry = copy_geometry
        self.kwargs = kwargs

    def __call__(self, image: sitk.Image) -> sitk.Image:
        """ImageFunction callable object:
        Process an image based on a given function.

        Parameters
        ----------
        image
            sitk.Image object to be processed.

        Returns
        -------
        sitk.Image
            The image processed with the given function.
        """

        result = self.function(image, **self.kwargs)
        if self.copy_geometry:
            result.CopyInformation(image)
        return result


class ArrayFunction(BaseOp):
    """ArrayFunction operation class:
    A callable class that takes in a function to be used to process an image from numpy array,
    and executes it.

    To instantiate:
        obj = ArrayFunction(function, copy_geometry, **kwargs)
    To call:
        result = obj(image)

    Parameters
    ----------
    function
        A function to be used for image processing.
        This function needs to have the following signature:
        - function(image: sitk.Image, **args)
        - The first argument needs to be an sitkImage, followed by optional arguments.

    copy_geometry, optional
        An optional argument to specify whether information about the image should be copied to the
        resulting image. Set to be true as a default.

    kwargs, optional
        Any number of arguements used in the given function.
    """

    def __init__(
        self, function: Function, copy_geometry: bool = True, **kwargs: Optional[Any]
    ):
        self.function = function
        self.copy_geometry = copy_geometry
        self.kwargs = kwargs

    def __call__(self, image: sitk.Image) -> sitk.Image:
        """ArrayFunction callable object:
        Processes an image from numpy array.

        Parameters
        ----------
        image
            sitk.Image object to be processed.

        Returns
        -------
        sitk.Image
            The image processed with a given function.
        """

        array, origin, direction, spacing = image_to_array(image)
        result = self.function(array, **self.kwargs)
        if self.copy_geometry:
            result = array_to_image(result, origin, direction, spacing)
        else:
            result = array_to_image(result)
        return result


# Segmentation ops


class StructureSetToSegmentation(BaseOp):
    """StructureSetToSegmentation operation class:

    A callable class that accepts ROI names, a StructureSet object, and a
    reference image, and returns a Segmentation mask.

    To instantiate:
        obj = StructureSetToSegmentation(roi_names)

    To call:
        mask = obj(structure_set, reference_image)

    Parameters
    ----------
    roi_names : Union[str, List[str], Dict[str, Union[str, List[str]]], None]
        ROI names or patterns to convert to segmentation:
        - `None` (default): All ROIs will be loaded
        - `str`: A single pattern (regex) to match ROI names.
        - `List[str]`: A list of patterns where each matches ROI names.
        - `Dict[str, str | List[str]]`: A dictionary where each key maps to a
          pattern (or list of patterns). The matched names are grouped under
          the same label.
        Both full names and case-insensitive regular expressions are allowed.
    continuous : bool, default=True
        Flag passed to 'physical_points_to_idxs' in 'StructureSet.to_segmentation'.
        Resolves errors caused by ContinuousIndex > Index.

    Notes
    -----
    If `roi_names` contains lists of strings, each matching
    name within a sublist will be assigned the same label. This means
    that `roi_names=['pat']` and `roi_names=[['pat']]` can lead
    to different label assignments, depending on how many ROI names
    match the pattern. E.g. if `self.roi_names = ['fooa', 'foob']`,
    passing `roi_names=['foo(a|b)']` will result in a segmentation with
    two labels, but passing `roi_names=[['foo(a|b)']]` will result in
    one label for both `'fooa'` and `'foob'`.

    If `roi_names` is kept empty ([]), the pipeline will process all ROIs/contours
    found according to their original names.

    In general, the exact ordering of the returned labels cannot be
    guaranteed (unless all patterns in `roi_names` can only match
    a single name or are lists of strings).

    """

    def __init__(
        self,
        roi_names: Union[str, List[str], Dict[str, Union[str, List[str]]], None] = None,
        continuous: bool = True,
    ):
        """Initialize the op."""
        self.roi_names = roi_names
        self.continuous = continuous

    def __call__(
        self,
        structure_set: StructureSet,
        reference_image: sitk.Image,
        existing_roi_indices: Dict[str, int],
        ignore_missing_regex: bool,
        roi_select_first: bool = False,
        roi_separate: bool = False,
    ) -> Segmentation:
        """Convert the structure set to a Segmentation object.

        Parameters
        ----------
        structure_set
            The structure set to convert.
        reference_image
            Image used as reference geometry.

        Returns
        -------
        Segmentation
            The segmentation object.
        """
        return structure_set.to_segmentation(
            reference_image,
            roi_names=self.roi_names,
            continuous=self.continuous,
            existing_roi_indices=existing_roi_indices,
            ignore_missing_regex=ignore_missing_regex,
            roi_select_first=roi_select_first,
            roi_separate=roi_separate,
        )


class FilterSegmentation:
    """FilterSegmentation operation class:
    A callable class that accepts ROI names, a Segmentation mask with all labels
    and returns only the desired Segmentation masks based on accepted ROI names.

    To instantiate:
        obj = StructureSet(roi_names)
    To call:
        mask = obj(structure_set, reference_image)

    Parameters
    ----------
    roi_names
        List of Region of Interests
    """

    def __init__(self, roi_patterns: Dict[str, str], continuous: bool = False):
        """Initialize the op.

        Parameters
        ----------
        roi_names
            List of ROI names to export. Both full names and
            case-insensitive regular expressions are allowed.
            All labels within one sublist will be assigned
            the same label.

        """
        self.roi_patterns = roi_patterns
        self.continuous = continuous

    def _assign_labels(
        self, names, roi_select_first: bool = False, roi_separate: bool = False
    ):
        """
        Parameters
        ----
        roi_select_first
            Select the first matching ROI/regex for each OAR, no duplicate matches.

        roi_separate
            Process each matching ROI/regex as individual masks, instead of consolidating into one mask
            Each mask will be named ROI_n, where n is the nth regex/name/string.
        """
        labels = {}
        cur_label = 0
        if names == self.roi_patterns:
            for i, name in enumerate(self.roi_patterns):
                labels[name] = i
        else:
            for _, pattern in enumerate(names):
                if sorted(names) == sorted(
                    list(labels.keys())
                ):  # checks if all ROIs have already been processed.
                    break
                if isinstance(pattern, str):
                    for i, name in enumerate(self.roi_names):
                        if re.fullmatch(pattern, name, flags=re.IGNORECASE):
                            labels[name] = cur_label
                            cur_label += 1
                else:  # if multiple regex/names to match
                    matched = False
                    for subpattern in pattern:
                        if roi_select_first and matched:
                            break  # break if roi_select_first and we're matched
                        for n, name in enumerate(self.roi_names):
                            if re.fullmatch(subpattern, name, flags=re.IGNORECASE):
                                matched = True
                                if not roi_separate:
                                    labels[name] = cur_label
                                else:
                                    labels[f"{name}_{n}"] = cur_label

                    cur_label += 1
        return labels

    def get_mask(self, reference_image, seg, mask, label, idx, continuous):
        size = seg.GetSize()
        seg_arr = sitk.GetArrayFromImage(seg)
        if len(size) == 5:
            size = size[:-1]
        elif len(size) == 3:
            size = size.append(1)

        idx_seg = (
            self.roi_names[label] - 1
        )  # SegmentSequence numbering starts at 1 instead of 0
        if (
            size[:-1] == reference_image.GetSize()
        ):  # Assumes `size` is length of 4: (x, y, z, channels)
            mask[:, :, :, idx] += seg[:, :, :, idx_seg]
        else:  # if 2D segmentations on 3D images
            frame = seg.frame_groups[idx_seg]
            ref_uid = (
                frame.DerivationImageSequence[0]
                .SourceImageSequence[0]
                .ReferencedSOPInstanceUID
            )  # unused but references InstanceUID of slice
            assert ref_uid is not None, "There was no ref_uid"  # dodging linter

            frame_coords = np.array(frame.PlanePositionSequence[0].ImagePositionPatient)
            img_coords = physical_points_to_idxs(
                reference_image, np.expand_dims(frame_coords, (0, 1))
            )[0][0]
            z = img_coords[0]

            mask[z, :, :, idx] += seg_arr[0, idx_seg, :, :]

    def __call__(
        self,
        reference_image: sitk.Image,
        seg: Segmentation,
        existing_roi_indices: Dict[str, int],
        ignore_missing_regex: bool = False,
        roi_select_first: bool = False,
        roi_separate: bool = False,
    ) -> Segmentation:
        """Convert the structure set to a Segmentation object.

        Parameters
        ----------
        structure_set
            The structure set to convert.
        reference_image
            Image used as reference geometry.

        Returns
        -------
        Segmentation
            The segmentation object.
        """
        from itertools import groupby

        # variable name isn't ideal, but follows StructureSet.to_segmentation convention
        self.roi_names = seg.raw_roi_names
        labels = {}

        # `roi_names` in .to_segmentation() method = self.roi_patterns
        if self.roi_patterns is None or self.roi_patterns == {}:
            self.roi_patterns = self.roi_names
            labels = self._assign_labels(
                self.roi_patterns, roi_select_first, roi_separate
            )  # only the ones that match the regex
        elif isinstance(self.roi_patterns, dict):
            for name, pattern in self.roi_patterns.items():
                if isinstance(pattern, str):
                    matching_names = list(
                        self._assign_labels([pattern], roi_select_first).keys()
                    )
                    if matching_names:
                        labels[name] = (
                            matching_names  # {"GTV": ["GTV1", "GTV2"]} is the result of _assign_labels()
                        )
                elif isinstance(
                    pattern, list
                ):  # for inputs that have multiple patterns for the input, e.g. {"GTV": ["GTV.*", "HTVI.*"]}
                    labels[name] = []
                    for pattern_one in pattern:
                        matching_names = list(
                            self._assign_labels([pattern_one], roi_select_first).keys()
                        )
                        if matching_names:
                            labels[name].extend(
                                matching_names
                            )  # {"GTV": ["GTV1", "GTV2"]}
        elif isinstance(
            self.roi_patterns, list
        ):  # won't this always trigger after the previous?
            labels = self._assign_labels(self.roi_patterns, roi_select_first)
        else:
            raise ValueError(f"{self.roi_patterns} not expected datatype")
        logger.debug(f"Found {len(labels)} labels", labels=labels)

        # removing empty labels from dictionary to prevent processing empty masks
        all_empty = True
        for v in labels.values():
            if v != []:
                all_empty = False
        if all_empty:
            if not ignore_missing_regex:
                raise ValueError(
                    f"No ROIs matching {self.roi_patterns} found in {self.roi_names}."
                )
            else:
                return None

        labels = {k: v for (k, v) in labels.items() if v != []}
        size = reference_image.GetSize()[::-1] + (len(labels),)
        mask = np.zeros(size, dtype=np.uint8)

        seg_roi_indices = {}
        if self.roi_patterns != {} and isinstance(self.roi_patterns, dict):
            for i, (name, label_list) in enumerate(labels.items()):
                for label in label_list:
                    self.get_mask(reference_image, seg, mask, label, i, self.continuous)
                seg_roi_indices[name] = i
        else:
            for name, label in labels.items():
                self.get_mask(reference_image, seg, mask, name, label, self.continuous)
            seg_roi_indices = {
                "_".join(k): v for v, k in groupby(labels, key=lambda x: labels[x])
            }

        mask[mask > 1] = 1
        mask = sitk.GetImageFromArray(mask, isVector=True)
        mask.CopyInformation(reference_image)
        return Segmentation(
            mask,
            roi_indices=seg_roi_indices,
            existing_roi_indices=existing_roi_indices,
            raw_roi_names=labels,
        )


class MapOverLabels(BaseOp):
    """MapOverLabels operation class:

    To instantiate:
        obj = MapOverLabels(op, include_background, return_segmentation)
    To call:
        mask = obj(segmentation, **kwargs)

    Parameters
    ----------
    op
        A processing function to be used for the operation.

    """

    def __init__(
        self, op, include_background: bool = False, return_segmentation: bool = True
    ):
        self.op = op
        self.include_background = include_background
        self.return_seg = return_segmentation

    def __call__(
        self, segmentation: Segmentation, **kwargs: Optional[Any]
    ) -> Segmentation:
        """MapOverLabels callable object:

        Parameters
        ----------
        include_background
            Specify whether to include background. Set to be false as a default.

        return_segmentation
            Specify whether to return segmentation. Set to be true as a default.

        **kwargs
            Arguments used for the processing function.

        Returns
        -------
        Segmentation
            The segmentation mask.
        """

        return map_over_labels(
            segmentation,
            self.op,
            include_background=self.include_background,
            return_segmentation=self.return_seg,
            **kwargs,
        )



class DirectoryNotFoundError(Exception):
    pass


class InvalidDicomError(Exception):
    pass


class NotRTSTRUCTError(InvalidDicomError):
    """Exception raised for errors in the RTSTRUCT DICOM file."""

    pass


class RTSTRUCTAttributeError(InvalidDicomError):
    """Exception raised for attribute errors in the RTSTRUCT DICOM file."""

    pass

class MissingROIError(KeyError):
    """Custom exception for missing ROI in the structure set."""

    pass


class ROIContourError(Exception):
    """Custom exception for missing ROI contour data in the RTSTRUCT file."""

    pass



import re
from concurrent.futures import ProcessPoolExecutor, as_completed
from os.path import commonpath
from pathlib import Path
from typing import Dict, List, Pattern, Set

from rich import progress
from rich.text import Text
from rich.tree import Tree

from imgtools.dicom import similar_tags, tag_exists
from imgtools.dicom.sort import (
    FileAction,
    InvalidDICOMKeyError,
    SorterBase,
    handle_file,
    resolve_path,
)

DEFAULT_PATTERN_PARSER: Pattern = re.compile(r"%([A-Za-z]+)|\{([A-Za-z]+)\}")


class DICOMSorter(SorterBase):
    """A specialized implementation of the `SorterBase` for sorting DICOM files by metadata.

    This class resolves paths for DICOM files based on specified
    target patterns, using metadata extracted from the files. The
    filename of each source file is preserved during this process.

    Attributes
    ----------
    source_directory : Path
        The directory containing the files to be sorted.
    logger : Logger
        The instance logger bound with the source directory context.
    dicom_files : list of Path
        The list of DICOM files found in the `source_directory`.
    format : str
        The parsed format string with placeholders for DICOM tags.
    keys : Set[str]
        DICOM tags extracted from the target pattern.
    invalid_keys : Set[str]
        DICOM tags from the pattern that are invalid.
    """

    def __init__(
        self,
        source_directory: Path,
        target_pattern: str,
        pattern_parser: Pattern = DEFAULT_PATTERN_PARSER,
    ) -> None:
        super().__init__(
            source_directory=source_directory,
            target_pattern=target_pattern,
            pattern_parser=pattern_parser,
        )
        self.logger.debug(
            "All DICOM Keys are Valid in target pattern", keys=self.keys
        )

    def validate_keys(self) -> None:
        """Validate the DICOM keys in the target pattern.

        If any invalid keys are found, it
        suggests similar valid keys and raises an error.
        """
        if not self.invalid_keys:
            return

        for key in sorted(self.invalid_keys):
            # TODO: keep this logic, but make the suggestion more user-friendly/readable
            similar = similar_tags(key)
            suggestion = (
                f"\n\tDid you mean: [bold green]{', '.join(similar)}[/bold green]?"
                if similar
                else " And [bold red]no similar keys[/bold red] found."
            )
            _error = (
                f"Invalid DICOM key: [bold red]{key}[/bold red].{suggestion}"
            )
            self._console.print(f"{_error}")
        self._console.print(f"Parsed Path: `{self.pattern_preview}`")
        errmsg = "Invalid DICOM Keys found."
        raise InvalidDICOMKeyError(errmsg)

    @property
    def invalid_keys(self) -> Set[str]:
        """Get the set of invalid keys.

        Essentially, this will check `pydicom.dictionary_has_tag` for each key
        in the pattern and return the set of keys that are invalid.

        Returns
        -------
        Set[str]
            The set of invalid keys.
        """
        return {key for key in self.keys if not tag_exists(key)}

    def execute(
        self,
        action: FileAction | str = FileAction.MOVE,
        overwrite: bool = False,
        dry_run: bool = False,
        num_workers: int = 1,
    ) -> None:
        """Execute the file action on DICOM files.

        Users are encouraged to use FileAction.HARDLINK for
        efficient storage and performance for large dataset, as well as
        protection against lost data.

        Using hard links can save disk space and improve performance by
        creating multiple directory entries (links) for a single file
        instead of duplicating the file content. This is particularly
        useful when working with large datasets, such as DICOM files,
        where storage efficiency is crucial.

        Parameters
        ----------
        action : FileAction, default: FileAction.MOVE
                The action to apply to the DICOM files (e.g., move, copy).
        overwrite : bool, default: False
                If True, overwrite existing files at the destination.
        dry_run : bool, default: False
                If True, perform a dry run without making any changes.
        num_workers : int, default: 1
                The number of worker threads to use for processing files.

        Raises
        ------
        ValueError
                If the provided action is not a valid FileAction.
        """
        if not isinstance(action, FileAction):
            action = FileAction.validate(action)

        self.logger.debug(
            f"Mapping {len(self.dicom_files)} files to new paths"
        )

        # Create a progress bar that can be used to track everything
        with self._progress_bar() as progress_bar:
            ################################################################################
            # Resolve new paths
            ################################################################################
            file_map: Dict[Path, Path] = self._resolve_new_paths(
                progress_bar=progress_bar, num_workers=num_workers
            )
        self.logger.info("Finished resolving paths")

        ################################################################################
        # Check if any of the resolved paths are duplicates
        ################################################################################
        file_map = self._check_duplicates(file_map)
        self.logger.info("Finished checking for duplicates")

        ################################################################################
        # Handle files
        ################################################################################
        if dry_run:
            self._dry_run(file_map)
            return

        with self._progress_bar() as progress_bar:
            task_files = progress_bar.add_task(
                "Handling files", total=len(file_map)
            )
            new_paths: List[Path | None] = []
            with ProcessPoolExecutor(max_workers=num_workers) as executor:
                future_to_file = {
                    executor.submit(
                        handle_file,
                        source_path,
                        resolved_path,
                        action,
                        overwrite,
                    ): source_path
                    for source_path, resolved_path in file_map.items()
                }
                for future in as_completed(future_to_file):
                    try:
                        result = future.result()
                        new_paths.append(result)
                        progress_bar.update(task_files, advance=1)
                    except Exception as e:
                        self.logger.exception(
                            "Failed to handle file",
                            exc_info=e,
                            file=future_to_file[future],
                        )

    def _check_duplicates(
        self, file_map: Dict[Path, Path]
    ) -> Dict[Path, Path]:
        """
        Check if any of the resolved paths are duplicates.

        Parameters
        ----------
        file_map : Dict[Path, Path]
            A dictionary mapping source paths to resolved paths.

        Returns
        -------
        Dict[Path, Path]
            A dictionary mapping source paths to resolved paths.

        Raises
        ------
        ValueError
            If any of the resolved paths are duplicates.
        """
        # opposite of the file_map
        # key: resolved path, value: list of source paths
        duplicate_paths: Dict[Path, List[Path]] = {}

        for source_path, resolved_path in file_map.items():
            if resolved_path in duplicate_paths:
                duplicate_paths[resolved_path].append(source_path)
            else:
                duplicate_paths[resolved_path] = [source_path]

        duplicates = False
        for resolved_path, source_paths in duplicate_paths.items():
            if len(source_paths) > 1:
                msg = f"Duplicate paths found for {resolved_path}: {source_paths}"
                self.logger.warning(msg)
                duplicates = True

        if duplicates:
            msg = "Duplicate paths found. Please check the log file for more information."
            raise ValueError(msg)

        return file_map

    def _resolve_new_paths(
        self, progress_bar: progress.Progress, num_workers: int = 1
    ) -> Dict[Path, Path]:
        """Resolve the new paths for all DICOM files using parallel processing.

        Parameters
        ----------
        progress_bar : progress.Progress
                Progress bar to use for tracking the progress of the operation.
        num_workers : int, default=1
                Number of threads to use for parallel processing.

        Returns
        -------
        Dict[Path, Path]
                A mapping of source paths to resolved paths.
        """
        task = progress_bar.add_task(
            "Resolving paths", total=len(self.dicom_files)
        )

        # Use ProcessPoolExecutor for parallel processing
        results: Dict[Path, Path] = {}
        with ProcessPoolExecutor(max_workers=num_workers) as executor:
            future_to_path = {
                executor.submit(
                    resolve_path, path, self.keys, self.format
                ): path
                for path in self.dicom_files
            }
            for future in as_completed(future_to_path):
                source, resolved = future.result()
                results[source] = resolved
                progress_bar.update(task, advance=1)

        return results

    def _dry_run(self, file_map: Dict[Path, Path]) -> None:
        """Perform a dry run without making any changes."""
        self._console.print(
            "[bold green]:double_exclamation_mark: Dry run mode enabled. No files will be moved or copied. :double_exclamation_mark: [/bold green]"
        )

        new_paths = sorted(list(file_map.values()))
        ppa = Path(self.pattern_preview).absolute()
        common_prefix: Path = self._common_prefix([ppa, *new_paths])
        common_prefix_styled = f"[bold yellow]{common_prefix}[/bold yellow]"
        self._console.print(
            f"\nCommon Prefix: :file_folder:{common_prefix_styled}\n\n"
        )

        tree = self._setup_tree(Path(common_prefix_styled))

        root = ppa.relative_to(common_prefix).as_posix()
        self._generate_tree_structure(
            root,
            tree,
        )
        self._build_tree(new_paths, tree, common_prefix)
        self._console.print(
            "[bold]Preview of the [magenta]parsed pattern[/magenta] and sample paths as directories:[/bold]\n"
        )
        self._console.print(tree)

    @staticmethod
    def _common_prefix(paths: List[Path]) -> Path:
        """Find the common prefix for a list of paths.

        Parameters
        ----------
        paths : List[Path]
            A list of paths.

        Returns
        -------
        str
            The common prefix of the paths.
        """
        return Path(
            commonpath(
                [str(path) for path in paths],
            ),
        )

    def _build_tree(
        self,
        paths: List[Path],
        tree: Tree,
        common_prefix: Path,
        max_children: int = 3,
    ) -> None:
        """
        Build a tree view given a set of paths (all having a common prefix).

        Parameters
        ----------
        paths : List[Path]
                        List of pathlib.Path objects representing the proposed paths.
        tree : Tree
                        A Rich Tree object where the directory structure will be added.
        common_prefix : Path
                        The common prefix for all paths. Used to normalize the input paths.
        """

        # Make all paths relative to the common prefix and sort them
        paths = [path.relative_to(common_prefix) for path in paths]
        # Create a dictionary to hold references to tree nodes
        node_map: Dict[Path, Tree] = {Path(): tree}  # Start with the root tree
        depth_counts: Dict[int, Set] = {
            0: set()
        }  # Track the number of nodes at each depth
        for path in paths:
            parts = list(path.parts)
            for depth, part in enumerate(parts):
                current_path = Path(*parts[: depth + 1])
                # Add the current path to the tree if it's not already in the map
                if current_path in node_map:
                    continue

                # add depth to the depth_counts
                depth_counts.setdefault(depth, set()).add(part)
                parent_path = Path(*parts[:depth])  # Determine parent path
                parent_node = node_map[parent_path]  # Get the parent Tree node

                if len(parent_node.children) < max_children:
                    # Add the current node to the tree
                    node_map[current_path] = parent_node.add(part)
                    break
                # Check if the parent has 3+ children already
                # If no "..." placeholder exists, add it and replace the 3rd child
                if not any(
                    child.label == "..." for child in parent_node.children
                ):
                    third_child = parent_node.children.pop(
                        2
                    )  # Remove the 3rd child
                    parent_node.add("...")  # Add the "..." placeholder
                    parent_node.add(third_child)  # Add the 3rd child back

                # Replace the last child after the "..." placeholder
                _ = parent_node.children.pop()  # Remove the last child
                node_map[current_path] = parent_node.add(part)
                break

        # Add counts to the first child and its children recursively
        def add_counts(node: Tree, depth: int) -> None:
            if depth in depth_counts:
                node.label = Text.assemble(
                    node.label,  # type: ignore
                    Text(
                        f" ({len(depth_counts[depth])} unique)",
                        style="bold green",
                    ),
                )
            for child in node.children:
                add_counts(child, depth + 1)

        add_counts(tree.children[0], 0)



from __future__ import annotations

import asyncio
import concurrent.futures
import os
import shutil
import tarfile
import zipfile
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Optional

import aiohttp
from rich import print  # noqa
from rich.console import Console
from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    TaskProgressColumn,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
)

# create a single console for all progress bars, so they don't clutter the output
console = Console()

try:
    from github import Github  # type: ignore # noqa
    from github.Repository import Repository  # type: ignore # noqa
except ImportError as e:
    raise ImportError(
        "PyGithub is required for the test data feature of med-imagetools. "
        "Install it using 'pip install med-imagetools[test]'."
    ) from e


@dataclass
class GitHubReleaseAsset:
    """
    Represents an asset in a GitHub release.

    Attributes
    ----------
    name : str
        Name of the asset (e.g., 'dataset.zip').
    url : str
        Direct download URL for the asset.
    content_type : str
        MIME type of the asset (e.g., 'application/zip').
    size : int
        Size of the asset in bytes.
    download_count : int
        Number of times the asset has been downloaded.
    """

    name: str
    url: str
    content_type: str
    size: int
    download_count: int


@dataclass
class GitHubRelease:
    """
    Represents a GitHub release.

    Attributes
    ----------
    tag_name : str
        The Git tag associated with the release.
    name : str
        The name of the release.
    body : str
        Release notes or description.
    html_url : str
        URL to view the release on GitHub.
    created_at : str
        ISO 8601 timestamp of release creation.
    published_at : str
        ISO 8601 timestamp of release publication.
    assets : List[GitHubReleaseAsset]
        List of assets in the release.
    """

    tag_name: str
    name: str
    body: str = field(repr=False)
    html_url: str
    created_at: str
    published_at: str
    assets: List[GitHubReleaseAsset]


async def download_dataset(
    download_link: str,
    file_path: Path,
    progress: Progress,
    timeout_seconds: int = 3600,
) -> Path:
    """
    Download a single dataset.

    Parameters
    ----------
    download_link : str
        The URL to download the dataset from.
    file_path : Path
        The path where the downloaded file will be saved.
    progress : Progress
        The progress bar to use for the download.
    timeout_seconds : int, optional
        The timeout for the download in seconds, by default 3600.

    Returns
    -------
    Path
        The path to the downloaded file.
    """
    timeout = aiohttp.ClientTimeout(total=timeout_seconds)
    async with aiohttp.ClientSession(timeout=timeout) as session:
        try:
            async with session.get(download_link) as response:
                total = int(response.headers.get("content-length", 0))
                task = progress.add_task(
                    f"[cyan]{file_path.name}...", total=total
                )
                with file_path.open("wb") as f:
                    async for chunk in response.content.iter_chunked(8192):
                        f.write(chunk)
                        progress.update(task, advance=len(chunk))
        except asyncio.TimeoutError:
            console.print(
                f"[bold red]Timeout while downloading "
                f"{file_path.name}. Please try again later.[/]"
            )
            raise
    return file_path


@dataclass
class MedImageTestData:
    """
    Manager for downloading and extracting med-image test data from GitHub releases.

    Attributes
    ----------
    repo_name : str
        The full name of the GitHub repository (e.g., 'user/repo').
    token : str | None
        Optional GitHub token for authenticated requests (higher rate limits).
    """

    repo_name: str = "bhklab/med-image_test-data"
    github: Github = field(init=False)
    repo: Repository = field(init=False)
    latest_release: GitHubRelease = field(init=False)
    downloaded_paths: List[Path] = field(default_factory=list, init=False)
    progress: Progress = field(
        default_factory=Progress, init=False, repr=False
    )

    # github parameters
    timeout = 300

    def __post_init__(self) -> None:
        token = os.environ.get("GITHUB_TOKEN", os.environ.get("GH_TOKEN"))
        if token:
            console.log("Using GH token")
            self.github = Github(token, timeout=self.timeout)
        else:
            console.log("No GH token found")
            self.github = Github(timeout=self.timeout)

        self.repo = self.github.get_repo(self.repo_name)
        self.get_latest_release()

    def get_release(self, tag: str = "latest") -> GitHubRelease:
        """
        Fetches the details of a specific release or the latest release if no tag is provided.

        Parameters
        ----------
        tag : str, optional
            The tag of the release to fetch. If 'latest', fetches the latest release.

        Returns
        -------
        GitHubRelease
            The details of the requested release.
        """
        if tag == "latest":
            release = self.repo.get_latest_release()
        else:
            release = self.repo.get_release(tag)

        assets = [
            GitHubReleaseAsset(
                name=asset.name,
                url=asset.browser_download_url,
                content_type=asset.content_type,
                size=asset.size,
                download_count=asset.download_count,
            )
            for asset in release.get_assets()
        ]

        return GitHubRelease(
            tag_name=release.tag_name,
            name=release.title,
            body=release.body or "",
            html_url=release.html_url,
            created_at=release.created_at.isoformat(),
            published_at=release.published_at.isoformat(),
            assets=assets,
        )

    def get_latest_release(self) -> GitHubRelease:
        """Fetches the latest release details from the repository."""
        self.latest_release = self.get_release("latest")
        return self.latest_release

    @property
    def datasets(self) -> List[GitHubReleaseAsset]:
        return self.latest_release.assets

    @property
    def dataset_names(self) -> List[str]:
        return [asset.name for asset in self.datasets]

    async def _download_asset(
        self,
        session: aiohttp.ClientSession,
        asset: GitHubReleaseAsset,
        dest: Path,
        progress: Progress,
        force: bool = False,
    ) -> Path:
        """
        Helper method to download a single asset asynchronously with a progress bar.

        Parameters
        ----------
        session : aiohttp.ClientSession
            The aiohttp session to use for the request.
        asset : GitHubReleaseAsset
            The asset to download.
        dest : Path
            Destination directory where the file will be saved.
        progress : Progress
            The progress bar to use for the download.

        Returns
        -------
        Path
            Path to the downloaded file.
        """
        dest.mkdir(parents=True, exist_ok=True)
        filepath = dest / asset.name
        # also check for name without "tar.gz" extension
        alternate_name = filepath.with_suffix("").with_suffix("")
        if filepath.exists() or alternate_name.exists():
            console.print(
                f"File {asset.name} already exists. Skipping download."
            )
            return filepath

        return await download_dataset(asset.url, filepath, progress)

    async def _download(
        self,
        dest: Path,
        assets: List[GitHubReleaseAsset],
        progress: Progress,
        force: bool = False,
    ) -> List[Path]:
        """
        Download specified assets or all assets if none are specified.

        Parameters
        ----------
        dest : Path
            Destination directory where the files will be saved.
        assets : List[GitHubReleaseAsset], optional
            List of assets to download. If None, all assets will be downloaded.
        progress : Progress
            The progress bar to use for the download.

        Returns
        -------
        List[Path]
            List of paths to the downloaded files.
        """

        async with aiohttp.ClientSession() as session:
            tasks = [
                self._download_asset(session, asset, dest, progress, force)
                for asset in assets
            ]
            self.downloaded_paths = await asyncio.gather(*tasks)
        return self.downloaded_paths

    def download(
        self,
        dest: Path,
        assets: Optional[List[GitHubReleaseAsset]] = None,
        exclude: Optional[List[str]] = None,
        force: bool = False,
        cores: Optional[int] = None,
    ) -> List[Path]:
        """
        Download specified assets synchronously, extracts by default.

        Parameters
        ----------
        dest : Path
            Destination directory where the files will be saved.
        exclude: List[str], optional
            List of assets to exclude from download.
            Can be patterns or exact names.
        assets : List[GitHubReleaseAsset], optional
            List of assets to download. If None, all assets will be downloaded

        Returns
        -------
        List[Path]
            List of paths to the downloaded files or extracted directories.
        """
        if assets is None:
            assets = self.latest_release.assets

        if exclude:
            import re

            console.log(f"Excluding assets on patterns: {', '.join(exclude)}")
            # try matching OR finding exact names
            exclude = [re.compile(f".*{name}.*") for name in exclude]
            exclude_assets = [
                asset
                for asset in assets
                if any(pattern.match(asset.name) for pattern in exclude)
            ]
            console.log(
                f"Excluding assets: {', '.join(asset.name for asset in exclude_assets)}"
            )
            assets = [asset for asset in assets if asset not in exclude_assets]

        console.print(f"Downloading assets to {dest.absolute()}...")
        with Progress(
            TextColumn("[bold blue]{task.description}", justify="right"),
            BarColumn(),
            TaskProgressColumn(),
            DownloadColumn(),
            TimeRemainingColumn(),
            TimeElapsedColumn(),
            transient=True,
            console=console,
        ) as progress:
            _ = asyncio.run(self._download(dest, assets, progress, force))

        extracted_paths = self.extract(force=force, cores=cores)
        for tar_file in self.downloaded_paths:
            if tar_file.exists():
                console.log(f"Removing downloaded file: {tar_file}")
                tar_file.unlink()
        return extracted_paths

    def extract(
        self, force: bool = False, cores: Optional[int] = None
    ) -> List[Path]:
        """Extract downloaded archives to the specified directory in parallel.

        Parameters
        ----------
        force : bool, optional
            Whether to force extraction by removing existing directories, by default False.
        cores : int, optional
            Number of cores to use for parallel extraction, by default uses all available cores.

        Returns
        -------
        List[Path]
            List of paths to the extracted directories.
        """
        if not self.downloaded_paths:
            raise ValueError(
                "No archives have been downloaded yet. Call `download` first."
            )

        if cores is None:
            cores = os.cpu_count()

        def extract_archive(path: Path) -> List[Path]:
            extract_path = path.with_suffix("").with_suffix("")
            extracted_paths: List[Path] = []

            if extract_path.exists():
                if force:
                    console.print(
                        f"Removing existing directory {extract_path}..."
                    )
                    shutil.rmtree(extract_path)
                else:
                    console.print(
                        f"Directory {extract_path} already"
                        " exists. Skipping extraction."
                    )
                    extracted_paths.append(extract_path)
                    return extracted_paths

            console.log(f"Extracting {path.name}...")
            if tarfile.is_tarfile(path):
                with tarfile.open(path, "r:*") as archive:
                    archive.extractall(extract_path.parent, filter="data")
                    extracted_paths.append(extract_path)
            elif zipfile.is_zipfile(path):
                with zipfile.ZipFile(path, "r") as archive:
                    archive.extractall(extract_path.parent)
                    extracted_paths.append(extract_path)
            else:
                console.print(f"Unsupported archive format: {path.name}")
            return extracted_paths

        with concurrent.futures.ThreadPoolExecutor(
            max_workers=cores
        ) as executor:
            results = list(
                executor.map(extract_archive, self.downloaded_paths)
            )

        # Flatten the list of lists
        return [item for sublist in results for item in sublist]


# Usage example
if __name__ == "__main__":  # pragma: no cover
    manager = MedImageTestData()

    print(manager)

    old_release = manager.get_latest_release()

    chosen_assets = old_release.assets[8:]

    dest_dir = Path("data")
    downloaded_files = manager.download(dest_dir, assets=chosen_assets)
    console.print(downloaded_files)



from .datagraph import DataGraph
from .dose import Dose
from .pet import PET
from .scan import Scan
from .segmentation import Segmentation, accepts_segmentations, map_over_labels
from .sparsemask import SparseMask
from .structureset import StructureSet
from .structureset_helpers import (
    DicomInput,
    extract_roi_meta,
    extract_roi_names,
    load_rtstruct_dcm,
    rtstruct_reference_uids,
)

__all__ = [
    "Segmentation",
    "map_over_labels",
    "accepts_segmentations",
    "StructureSet",
    "PET",
    "Dose",
    "DataGraph",
    "SparseMask",
    "Scan",
    # structureset_helpers
    "DicomInput",
    "extract_roi_meta",
    "extract_roi_names",
    "load_rtstruct_dcm",
    "rtstruct_reference_uids",
]



__version__ = "1.20.0"

from .coretypes import Coordinate3D, Size3D, Spacing3D, Vector3D
from .datasets import example_data, example_data_paths
from .dicom import find_dicoms, lookup_tag, similar_tags, tag_exists
from .logging import logger

__all__ = [
    "find_dicoms",
    "lookup_tag",
    "similar_tags",
    "tag_exists",
    "logger",
    ## coretypes
    "Coordinate3D",
    "Size3D",
    "Spacing3D",
    "Vector3D",
    ## example data
    "example_data",
    "example_data_paths",
]



from .pattern_resolver import (
    MissingPlaceholderValueError,
    PatternResolver,
    PatternResolverError,
)

__all__ = ["PatternResolver", "PatternResolverError", "MissingPlaceholderValueError"]



import os
import pathlib
import shutil
import warnings
from multiprocessing import cpu_count

import numpy as np
import pytest
import SimpleITK as sitk

from imgtools.io import ImageFileLoader, ImageFileWriter
from imgtools.ops import (
    BaseInput as Input,
    BaseOutput as Output,
)
from imgtools.pipeline import Pipeline


@pytest.fixture
def sample_input_output(tmp_path):
    input_paths = []
    output_paths = []
    for i in range(2):
        input_path = tmp_path / f'input_{i}'
        output_path = tmp_path / f'output_{i}'
        input_path.mkdir(exist_ok=True)
        output_path.mkdir(exist_ok=True)
        input_paths.append(input_path)
        output_paths.append(output_path)

        # generate some test data
        test_inputs = [sitk.GetImageFromArray(np.random.random((10, 10, 2))) for _ in range(4)]
        for j, img in enumerate(test_inputs):
            path = input_path / f'test{j}.nrrd'
            sitk.WriteImage(img, str(path))

    yield input_paths, output_paths
    # clean up
    shutil.rmtree(tmp_path)


class SimplePipelineTest(Pipeline):
    def __init__(self, input_path, output_path, n_jobs) -> None:
        super().__init__(n_jobs=n_jobs, show_progress=False)
        self.input_path = input_path
        self.output_path = output_path
        self.image_input = Input(ImageFileLoader(self.input_path))
        self.image_output = Output(ImageFileWriter(self.output_path))

    def process_one_subject(self, subject_id) -> None:
        image = self.image_input(subject_id)
        self.image_output(subject_id, image)


@pytest.mark.parametrize('n_jobs', [1, 2])
def test_output(n_jobs, sample_input_output) -> None:
    if cpu_count() < 2 and n_jobs == 2:
        n_jobs = 0
    input_paths, output_paths = sample_input_output
    pipeline = SimplePipelineTest(input_paths[0], output_paths[0], n_jobs)
    pipeline.run()

    input_dir, output_dir = input_paths[0], output_paths[0]

    for input_file, output_file in zip(
        sorted(os.listdir(input_dir)), sorted(os.listdir(output_dir))
    ):
        assert os.path.exists(pathlib.Path(output_dir, output_file))
        test_output = sitk.GetArrayFromImage(
            sitk.ReadImage(pathlib.Path(output_dir, output_file).as_posix())
        )
        true_output = sitk.GetArrayFromImage(
            sitk.ReadImage(pathlib.Path(input_dir, input_file).as_posix())
        )
        assert np.allclose(test_output, true_output)


class MultiInputPipelineTest(Pipeline):
    def __init__(
        self, input_path_0, input_path_1, output_path_0, output_path_1, n_jobs, missing_strategy
    ) -> None:
        super().__init__(n_jobs=n_jobs, missing_strategy=missing_strategy, show_progress=False)
        self.input_path_0 = input_path_0
        self.input_path_1 = input_path_1
        self.output_path_0 = output_path_0
        self.output_path_1 = output_path_1
        self.image_input_0 = Input(ImageFileLoader(self.input_path_0))
        self.image_input_1 = Input(ImageFileLoader(self.input_path_1))
        self.image_output_0 = Output(ImageFileWriter(self.output_path_0))
        self.image_output_1 = Output(ImageFileWriter(self.output_path_1))

    def process_one_subject(self, subject_id) -> None:
        image_0 = self.image_input_0(subject_id)
        image_1 = self.image_input_1(subject_id)
        if image_0 is not None:
            self.image_output_0(subject_id, image_0)
        if image_1 is not None:
            self.image_output_1(subject_id, image_1)


@pytest.mark.parametrize('n_jobs', [1, 2])
@pytest.mark.parametrize('missing_strategy', ['pass', 'drop'])
def test_missing_handling(n_jobs, missing_strategy, sample_input_output) -> None:
    if cpu_count() < 2 and n_jobs == 2:
        n_jobs = 0
    input_paths, output_paths = sample_input_output
    # simulate partial missing data
    os.remove(pathlib.Path(input_paths[0], 'test0.nrrd').as_posix())

    pipeline = MultiInputPipelineTest(
        input_paths[0], input_paths[1], output_paths[0], output_paths[1], n_jobs, missing_strategy
    )
    with warnings.catch_warnings(record=True) as w:
        pipeline.run()
        assert len(w) == 1
        assert missing_strategy in str(w[-1].message)

    if missing_strategy == 'drop':
        assert all(
            [
                not os.path.exists(pathlib.Path(output_paths[0], 'test0.nrrd').as_posix()),
                not os.path.exists(pathlib.Path(output_paths[1], 'test0.nrrd').as_posix()),
            ]
        )
    else:
        assert all(
            [
                not os.path.exists(pathlib.Path(output_paths[0], 'test0.nii.gz').as_posix()),
                os.path.exists(pathlib.Path(output_paths[1], 'test0.nii.gz').as_posix()),
            ]
        )



import pathlib

import click

from imgtools.dicom.sort import DICOMSorter, FileAction
from imgtools.logging import logger

""" TODO: Implement overwrite option

Until we implement proper handling of overwrite behavior
The overwrite option could lead to data loss. Consider adding additional safety measures:

Require an additional confirmation when overwrite is enabled
Add a backup mechanism for overwritten files
Log all overwritten files for audit purposes

"""
DEFAULT_OVERWRITE_BEHAVIOR = False


@click.command()
@click.argument(
	'source_directory',
	type=click.Path(
		exists=True,
		file_okay=False,
		dir_okay=True,
		readable=True,
		path_type=pathlib.Path,
		resolve_path=True,
	),
)
@click.argument(
	'target_directory',
	type=click.Path(
		file_okay=False,
		writable=True,
		path_type=str,  # TODO: figure out if patterned path i.e data/%PatientID can be a pathlib.Path
	),
)
@click.option(
	'--action',
	'-a',
	type=click.Choice(FileAction.choices(), case_sensitive=False),
	required=True,
	help='Action to perform on the files.',
)
@click.option(
	'-n',
	'--dry-run',
	is_flag=True,
	help='Do not move or copy files, just print what would be done. Always recommended to use this first to confirm the operation!',
)
@click.option(
	'-j',
	'--num-workers',
	type=int,
	default=1,
	show_default=True,
	help='Number of worker processes to use for sorting.',
)
@click.help_option(
	'-h',
	'--help',
)
def dicomsort(
	source_directory: pathlib.Path,
	target_directory: str,
	action: FileAction,
	dry_run: bool,
	num_workers: int,
) -> None:
	"""Sorts DICOM files into directories based on their tags."""
	logger.info(f'Sorting DICOM files in {source_directory}.')
	logger.debug('Debug Args', args=locals())
	# TODO: eagerly validate target pattern somehow?

	try:
		sorter = DICOMSorter(
			source_directory=source_directory,
			target_pattern=target_directory,
		)
	except Exception as e:
		logger.exception('Failed to initialize DICOMSorter')
		raise click.Abort() from e

	sorter.execute(
		action=action,
		overwrite=DEFAULT_OVERWRITE_BEHAVIOR,
		dry_run=dry_run,
		num_workers=num_workers,
	)


if __name__ == '__main__':
	dicomsort()



from io import BytesIO
from pathlib import Path
from typing import TypeAlias

from pydicom import dcmread
from pydicom.dataset import FileDataset

from imgtools.exceptions import (
    InvalidDicomError,
    NotRTSTRUCTError,
    RTSTRUCTAttributeError,
)

# Define a type alias for DICOM input types
DicomInput: TypeAlias = FileDataset | str | Path | bytes


def load_rtstruct_dcm(
    rtstruct_input: DicomInput,
    force: bool = True,
    stop_before_pixels: bool = True,
) -> FileDataset:
    """Load an RTSTRUCT DICOM file and return the parsed FileDataset object.

    Parameters
    ----------
    rtstruct_input : FileDataset | str | Path | bytes
        Input DICOM file as a `pydicom.FileDataset`, file path, or byte stream.
    force : bool, optional
        Whether to allow reading DICOM files missing the *File Meta Information*
        header, by default True.
    stop_before_pixels : bool, optional
        Whether to stop reading the DICOM file before loading pixel data, by default True.

    Returns
    -------
    FileDataset
        Parsed RTSTRUCT DICOM dataset.

    Raises
    ------
    InvalidDicomError
        If the input is of an unsupported type or cannot be read as a DICOM file.
    NotRTSTRUCTError
        If the input file is not an RTSTRUCT (i.e., `Modality` field is not "RTSTRUCT").
    """

    match rtstruct_input:
        case FileDataset():
            dicom = rtstruct_input
        case str() | Path():
            dicom = dcmread(
                rtstruct_input,
                force=force,
                stop_before_pixels=stop_before_pixels,
            )
        case bytes():
            dicom = dcmread(
                BytesIO(rtstruct_input),
                force=force,
                stop_before_pixels=stop_before_pixels,
            )
        case _:
            msg = (
                f"Invalid input type for 'rtstruct_input': {type(rtstruct_input)}. "
                "Must be a str, Path, or bytes object."
            )
            raise InvalidDicomError(msg)

    if dicom.Modality != "RTSTRUCT":
        msg = f"The provided DICOM is not an RTSTRUCT file. Found Modality: {dicom.Modality}"
        raise NotRTSTRUCTError(msg)

    return dicom


def extract_roi_meta(rtstruct: DicomInput) -> list[dict[str, str]]:
    """Extract ROI metadata from an RTSTRUCT DICOM file.

    Iterate over the `StructureSetROISequence` in the RTSTRUCT file and extract:
        - "ROINumber": Unique identifier for the ROI.
        - "ROIName": Name of the ROI.
        - "ROIGenerationAlgorithm": Algorithm used to generate the ROI.

    Parameters
    ----------
    rtstruct : FileDataset | str | Path | bytes
        Input RTSTRUCT DICOM dataset or file path.

    Returns
    -------
    list of dict[str, str]
        A list of dictionaries, each containing metadata for an ROI.

    Raises
    ------
    RTSTRUCTAttributeError
        If the RTSTRUCT file does not contain the required `StructureSetROISequence`.
    """

    dcm_rtstruct: FileDataset = load_rtstruct_dcm(rtstruct)
    try:
        roi_sequence = dcm_rtstruct.StructureSetROISequence
    except AttributeError as e:
        errmsg = "Failed to extract ROISequence from the RTSTRUCT file."
        raise RTSTRUCTAttributeError(errmsg) from e
    roi_metas = []
    for roi in roi_sequence:
        roi_meta = {}
        roi_meta["ROINumber"] = getattr(roi, "ROINumber", "")
        roi_meta["ROIName"] = getattr(roi, "ROIName", "")
        roi_meta["ROIGenerationAlgorithm"] = getattr(
            roi, "ROIGenerationAlgorithm", ""
        )
        roi_metas.append(roi_meta)
    return roi_metas


def extract_roi_names(rtstruct: DicomInput) -> list[str]:
    """Extract a list of ROI names from an RTSTRUCT DICOM file.

    Parameters
    ----------
    rtstruct : FileDataset | str | Path | bytes
        Input RTSTRUCT DICOM dataset or file path.

    Returns
    -------
    list of str
        A list of ROI names extracted from the RTSTRUCT file.

    Raises
    ------
    RTSTRUCTAttributeError
        If the RTSTRUCT file does not contain the required `StructureSetROISequence`.
    """
    dcm_rtstruct: FileDataset = load_rtstruct_dcm(rtstruct)
    try:
        roi_sequence = dcm_rtstruct.StructureSetROISequence
    except AttributeError as e:
        errmsg = "Failed to extract ROISequence from the RTSTRUCT file."
        raise RTSTRUCTAttributeError(errmsg) from e
    roi_names = [roi.ROIName for roi in roi_sequence]
    return roi_names


def rtstruct_reference_uids(rtstruct: DicomInput) -> tuple[str, str]:
    """Retrieve the referenced SeriesInstanceUID and StudyInstanceUID from an RTSTRUCT.

    Parameters
    ----------
    rtstruct : FileDataset | str | Path | bytes
        Input RTSTRUCT DICOM dataset or file path.

    Returns
    -------
    tuple[str, str]
        - Referenced `SeriesInstanceUID` (str).
        - Referenced `StudyInstanceUID` (str).

    Raises
    ------
    RTSTRUCTAttributeError
        If the RTSTRUCT file does not contain the required reference fields.
    """
    dcm_rtstruct: FileDataset = load_rtstruct_dcm(rtstruct)
    try:
        referenced_series_instance_uid = str(
            dcm_rtstruct.ReferencedFrameOfReferenceSequence[0]
            .RTReferencedStudySequence[0]
            .RTReferencedSeriesSequence[0]
            .SeriesInstanceUID
        )
        referenced_study_instance_uid = str(
            dcm_rtstruct.ReferencedFrameOfReferenceSequence[0]
            .RTReferencedStudySequence[0]
            .ReferencedSOPInstanceUID
        )
        return referenced_series_instance_uid, referenced_study_instance_uid
    except (AttributeError, IndexError) as e:
        errmsg = "Failed to extract Referenced SeriesInstanceUID or Referenced StudyInstanceUID from the RTSTRUCT file."
        raise RTSTRUCTAttributeError(errmsg) from e



from typing import List, Tuple

import numpy as np
import SimpleITK as sitk

# Define type aliases for better readability
Array3D = Tuple[float, float, float]
ImageArrayMetadata = Tuple[np.ndarray, Array3D, Array3D, Array3D]


def array_to_image(
    array: np.ndarray,
    origin: Array3D = (0.0, 0.0, 0.0),
    direction: Tuple[float, ...] = (
        1.0,
        0.0,
        0.0,
        0.0,
        1.0,
        0.0,
        0.0,
        0.0,
        1.0,
    ),
    spacing: Array3D = (1.0, 1.0, 1.0),
    reference_image: sitk.Image = None,
) -> sitk.Image:
    """Convert a numpy array to a SimpleITK image with optional metadata.

    Parameters
    ----------
    array : np.ndarray
        The numpy array to convert.
    origin : Array3D, optional
        The origin of the image (default is (0.0, 0.0, 0.0)).
    direction : Tuple[float, ...], optional
        The direction cosines of the image (default is identity matrix).
    spacing : Array3D, optional
        The pixel spacing of the image (default is (1.0, 1.0, 1.0)).
    reference_image : sitk.Image, optional
        A reference SimpleITK image to copy metadata from (default is None).

    Returns
    -------
    sitk.Image
        The resulting SimpleITK image.
    """
    image = sitk.GetImageFromArray(array)
    if reference_image is not None:
        image.CopyInformation(reference_image)
    else:
        image.SetOrigin(origin)
        image.SetDirection(direction)
        image.SetSpacing(spacing)

    return image


def image_to_array(image: sitk.Image) -> ImageArrayMetadata:
    """Convert a SimpleITK image to a numpy array along with its metadata.

    Parameters
    ----------
    image : sitk.Image
        The SimpleITK image to convert.

    Returns
    -------
    ImageArrayMetadata
        A tuple containing:
        - The image as a numpy array.
        - The origin of the image (tuple of floats).
        - The direction cosines of the image (tuple of floats).
        - The pixel spacing of the image (tuple of floats).
    """
    origin: Array3D = image.GetOrigin()
    direction: Array3D = image.GetDirection()
    spacing: Array3D = image.GetSpacing()
    array: np.ndarray = sitk.GetArrayFromImage(image)
    return array, origin, direction, spacing


def physical_points_to_idxs(
    image: sitk.Image,
    points: List[np.ndarray],
    continuous: bool = False,
) -> List[np.ndarray]:
    """Convert physical points to image indices based on the reference image's
    geometry.

    This function uses the geometry of a SimpleITK image (origin, spacing,
    direction) to convert real-world physical coordinates into indices in the
    image grid. It optionally supports continuous indices for sub-pixel
    precision.

    Parameters
    ----------
    image : sitk.Image
        The reference SimpleITK image.
    points : List[np.ndarray]
        List of 3D physical points to transform.
    continuous : bool, optional
        If True, returns continuous indices; otherwise, returns integer indices.
        Default is False.

    Returns
    -------
    List[np.ndarray]
        A list of transformed points in image index space, reversed to match
        library conventions.

    Notes
    -----
    The following steps occur within the function:
    1. A `numpy.vectorize` function is defined to apply the transformation
       method (physical to index) to each 3D point in the input array.
    2. The transformation is applied to each set of points in the list,
       reversing the coordinate order to match the library's indexing
       convention.
    """
    # Select the appropriate transformation function based on the `continuous` parameter.
    transform = (
        image.TransformPhysicalPointToContinuousIndex
        if continuous
        else image.TransformPhysicalPointToIndex
    )

    # Step 1: Define a vectorized transformation function
    # The lambda function takes a single 3D point `x` and:
    # - Applies the selected transformation (`transform(x)`) to convert it from physical space to index space.
    # - Wraps the result into a numpy array for further processing.
    # `np.vectorize` creates a vectorized function that can process arrays of points in one call.
    # The `signature="(3)->(3)"` ensures the transformation operates on 3D points, returning 3D results.
    vectorized_transform = np.vectorize(
        lambda x: np.array(transform(x)), signature="(3)->(3)"
    )

    # Step 2: Apply the vectorized transformation to all slices of points.
    # For each 2D array `slc` in the `points` list:
    # - `vectorized_transform(slc)` applies the transformation to all points in `slc`.
    # - `[:, ::-1]` reverses the coordinate order (from (x, y, z) to (z, y, x)) to match the library's convention.
    # The result is stored as a list of numpy arrays (`t_points`), each corresponding to a transformed slice.
    t_points: List[np.ndarray] = [
        vectorized_transform(slc)[:, ::-1] for slc in points
    ]

    # Return the list of transformed points.
    return t_points


def idxs_to_physical_points(image: sitk.Image, idxs: np.ndarray) -> np.ndarray:
    """
    Converts image indices to physical points based on the reference image's
    geometry.

    Parameters
    ----------
    image : sitk.Image
        The reference SimpleITK image.
    idxs : np.ndarray
        Array of 3D indices (continuous or discrete).

    Returns
    -------
    np.ndarray
        Physical coordinates corresponding to the given indices.
    """
    continuous = np.issubdtype(idxs.dtype, np.floating)
    transform = (
        image.TransformContinuousIndexToPhysicalPoint
        if continuous
        else image.TransformIndexToPhysicalPoint
    )
    vectorized_transform = np.vectorize(
        lambda x: np.array(transform(x)), signature="(3)->(3)"
    )
    return vectorized_transform(idxs)



from .old_writers import (
    BaseSubjectWriter,
    BaseWriter,
    HDF5Writer,
    ImageFileWriter,
    MetadataWriter,
    NumpyWriter,
    SegNrrdWriter,
)

__all__ = [
    "BaseWriter",
    "BaseSubjectWriter",
    "ImageFileWriter",
    "SegNrrdWriter",
    "NumpyWriter",
    "HDF5Writer",
    "MetadataWriter",
]



import re
import warnings

import pytest

from imgtools.dicom.sort.exceptions import InvalidPatternError
from imgtools.pattern_parser.parser import PatternParser

# Disable warnings for the purpose of this test
warnings.filterwarnings('ignore', category=pytest.PytestWarning)

# Define a mapping of pattern parsers to their valid test cases
# The second element of tuple is a list of test cases, each containing:
# - A pattern string
# - The expected formatted pattern
# - A list of expected keys
pattern_parsers_and_patterns = [
    # Default pattern parser
    (
        re.compile(r'%<(\w+)>|\{(\w+)\}'),
        [
            ('%<Key1> and {Key2}', '%(Key1)s and %(Key2)s', ['Key1', 'Key2']),
            ('%<Key1>-%<Key2>', '%(Key1)s-%(Key2)s', ['Key1', 'Key2']),
            ('{Key1}_{Key2}', '%(Key1)s_%(Key2)s', ['Key1', 'Key2']),
            ('{Key1} and some text', '%(Key1)s and some text', ['Key1']),
        ],
    ),
    # Alternative placeholder parser
    (
        re.compile(r'<(\w+)>|\[(\w+)\]'),
        [
            ('<Key1> and [Key2]', '%(Key1)s and %(Key2)s', ['Key1', 'Key2']),
            ('<Key1>-<Key2>', '%(Key1)s-%(Key2)s', ['Key1', 'Key2']),
            ('[Key1]_[Key2]', '%(Key1)s_%(Key2)s', ['Key1', 'Key2']),
            ('<Key1> and plain text', '%(Key1)s and plain text', ['Key1']),
        ],
    ),
    # Custom placeholder parser
    (
        re.compile(r'#(\w+)!|\$<(\w+)>'),
        [
            ('#Key1! and $<Key2>', '%(Key1)s and %(Key2)s', ['Key1', 'Key2']),
            ('#Key1!-$<Key2>', '%(Key1)s-%(Key2)s', ['Key1', 'Key2']),
            ('$<Key1>_#Key2!', '%(Key1)s_%(Key2)s', ['Key1', 'Key2']),
            ('$<Key1> and filler text', '%(Key1)s and filler text', ['Key1']),
        ],
    ),
    # Additional pattern example, no numbers involved in keys
    (
        re.compile(r'%([A-Za-z]+)|\{([A-Za-z]+)\}'),
        [
            ('%Key and {Keytwo}', '%(Key)s and %(Keytwo)s', ['Key', 'Keytwo']),
            ('%Key1 and {Keytwo}', '%(Key)s1 and %(Keytwo)s', ['Key', 'Keytwo']),
            ('%Key1-%Keytwo2', '%(Key)s1-%(Keytwo)s2', ['Key', 'Keytwo']),
            ('{Key}_{KeyTWO}', '%(Key)s_%(KeyTWO)s', ['Key', 'KeyTWO']),
            ('{Key} and some text', '%(Key)s and some text', ['Key']),
        ],
    ),
]


class TestPatternParser:
    @pytest.mark.parametrize('pattern_parser, test_cases', pattern_parsers_and_patterns)
    def test_valid_patterns(self, pattern_parser, test_cases) -> None:
        for pattern, expected_format, expected_keys in test_cases:
            parser = PatternParser(pattern, pattern_parser)
            formatted_pattern, keys = parser.parse()
            assert formatted_pattern == expected_format
            assert keys == expected_keys
            assert parser.keys == expected_keys

    @pytest.mark.parametrize('pattern_parser, _', pattern_parsers_and_patterns)
    def test_empty_pattern(self, pattern_parser, _) -> None:
        with pytest.raises(AssertionError, match='Pattern must be a non-empty string.'):
            parser = PatternParser('', pattern_parser)
            parser.parse()

    @pytest.mark.parametrize('pattern_parser, _', pattern_parsers_and_patterns)
    def test_invalid_pattern(self, pattern_parser, _) -> None:
        with pytest.raises(InvalidPatternError, match='Pattern must contain placeholders matching'):
            parser = PatternParser('Invalid pattern', pattern_parser)
            parser.parse()

    @pytest.mark.parametrize('pattern_parser, _', pattern_parsers_and_patterns)
    def test_none_pattern(self, pattern_parser, _) -> None:
        with pytest.raises(AssertionError, match='Pattern must be a non-empty string.'):
            parser = PatternParser(None, pattern_parser)
            parser.parse()



"""
This is a work in progress to break up the ops/ops.py file into smaller, more manageable pieces.
"""

from abc import ABC, abstractmethod
from typing import Any
from imgtools.io.loaders import BaseLoader
from imgtools.io.writers import BaseWriter


class BaseOp(ABC):
    """Abstract base class for operations.

    Classes inheriting from this must implement the `__call__` method.
    """

    @abstractmethod
    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """Perform the operation."""
        pass

    def __repr__(self) -> str:
        """
        Generate a string representation of the operation instance.

        Returns
        -------
        str
            The string representation of the object.
        """
        attrs = [(k, v) for k, v in self.__dict__.items() if not k.startswith("_")]
        attrs = [(k, f"'{v}'") if isinstance(v, str) else (k, v) for k, v in attrs]
        args = ", ".join(f"{k}={v}" for k, v in attrs)
        return f"{self.__class__.__name__}({args})"


class BaseInput(BaseOp):
    """Abstract base class for input operations.

    Parameters
    ----------
    loader : BaseLoader
        An instance of a subclass of `BaseLoader` responsible for loading input data.
    """

    _loader: BaseLoader

    def __init__(self, loader: BaseLoader):
        if not isinstance(loader, BaseLoader):
            msg = f"loader must be a subclass of io.BaseLoader, got {type(loader)}"
            raise ValueError(msg)
        self._loader = loader

    @abstractmethod
    def __call__(self, key: Any) -> Any:
        """Retrieve input data."""
        pass


class BaseOutput(BaseOp):
    """Abstract base class for output operations.

    Parameters
    ----------
    writer : BaseWriter
        An instance of a subclass of `BaseWriter` responsible for writing output data.
    """

    _writer: BaseWriter

    def __init__(self, writer: BaseWriter):
        if not isinstance(writer, BaseWriter):
            msg = f"writer must be a subclass of io.BaseWriter, got {type(writer)}"
            raise ValueError(msg)
        self._writer = writer

    @abstractmethod
    def __call__(self, key: Any, *args: Any, **kwargs: Any) -> None:
        """Write output data.

        Parameters
        ----------
        key : Any
            The identifier for the data to be written.
        *args : Any
            Positional arguments for the writing process.
        **kwargs : Any
            Keyword arguments for the writing process.
        """
        pass


if __name__ == "__main__":
    # Mock implementations
    class MockLoader(BaseLoader):
        def get(self, key: Any) -> str:
            return f"Data loaded for key: {key}"

    class MockWriter(BaseWriter):
        def put(self, key: Any, *args: Any, **kwargs: Any) -> None:
            print(f"Data written for key: {key} with args: {args} and kwargs: {kwargs}")

    # Concrete subclass of BaseInput
    class ExampleInput(BaseInput):
        def __call__(self, key: Any) -> str:
            return self._loader.get(key)

    # Concrete subclass of BaseOutput
    class ExampleOutput(BaseOutput):
        def __call__(self, key: Any, *args: Any, **kwargs: Any) -> None:
            self._writer.put(key, *args, **kwargs)

    # Demonstrating usage
    loader = MockLoader()
    writer = MockWriter(
        root_directory="example_dir",
        filename_format="example_{key}.txt",
        create_dirs=True,
    )

    example_input = ExampleInput(loader)
    example_output = ExampleOutput(writer)

    # Using ExampleInput
    key = "example_key"
    loaded_data = example_input(key)
    print("Loaded Data:", loaded_data)
    print("ExampleInput Repr:", repr(example_input))

    # Using ExampleOutput
    example_output(key, "example_data", extra_param="value")
    print("ExampleOutput Repr:", repr(example_output))



from importlib.resources import files

import SimpleITK as sitk

_data_path = files("imgtools.datasets.data")


def data_paths() -> dict[str, str]:
    """Return paths to example dataset files.

    Returns:
        dict[str, str]: A dictionary mapping dataset keys to their file paths.
        Available keys are:
        - "duck": Path to the merged duck with star image
        - "mask": Path to the star mask

    Raises:
        FileNotFoundError: If any of the dataset files are missing
    """
    return {
        "duck": str(_data_path.joinpath("merged_duck_with_star.nii.gz")),
        "mask": str(_data_path.joinpath("star_mask.nii.gz")),
    }


def data_images() -> dict[str, sitk.Image]:
    """Return example dataset images.

    Returns:
        dict[str, sitk.Image]: A dictionary mapping dataset keys to their SimpleITK images.
        Available keys are:
        - "duck": The merged duck with star image
        - "mask": The star mask
    """
    paths = data_paths()
    return {key: sitk.ReadImage(path) for key, path in paths.items()}



"""
Input classes refactored to use the BaseInput abstract base class.
"""

import pathlib
import time
from typing import Any, List, Optional

import pandas as pd

from imgtools.crawler import crawl
from imgtools.io.loaders import (
    ImageCSVLoader,
    ImageFileLoader,
    read_image,
)
from imgtools.logging import logger
from imgtools.modules.datagraph import DataGraph

from .base_classes import BaseInput


class CrawlGraphInput(BaseInput):
    """CrawlGraphInput class for loading and indexing datasets.

    This class crawls through the specified directory to index the dataset,
    creates a graph of the dataset, and allows querying the graph based on modalities.

    Parameters
    ----------
    dir_path : str | pathlib.Path
        Path to the dataset's top-level directory. The crawler/indexer will start at this directory.

    n_jobs : int
        Number of parallel jobs to run when crawling. Default is -1.

    update_crawl : bool
        Whether to update the crawled index. Default is False.

    imgtools_dir : str
        Directory name for imgtools output files. Default is ".imgtools".

    db : dict
        Dictionary containing the indexed dataset, created by the crawler.

    graph : DataGraph
        DataGraph object containing the dataset graph.
    """

    dir_path: str
    n_jobs: int
    update_crawl: bool
    imgtools_dir: str

    db: dict
    graph: DataGraph

    def __init__(
        self,
        dir_path: str | pathlib.Path,
        n_jobs: int = -1,
        update_crawl: bool = False,
        imgtools_dir: str = ".imgtools",
    ) -> None:
        self.dir_path = pathlib.Path(dir_path)
        self.dataset_name = self.dir_path.name
        self.n_jobs = n_jobs
        self.update_crawl = update_crawl
        self.imgtools_dir = imgtools_dir

        self.csv_path = self._create_path(f"imgtools_{self.dataset_name}.csv")
        self.json_path = self._create_path(
            f"imgtools_{self.dataset_name}.json"
        )
        self.edge_path = self._create_path(
            f"imgtools_{self.dataset_name}_edges.csv"
        )

        start = time.time()
        self.db = self._crawl()
        self.graph = self._init_graph()

        logger.info("Crawl and graph completed", time=time.time() - start)

    def _create_path(self, file_name: str) -> pathlib.Path:
        """Helper function to create paths.

        TODO: would be nice to allow users to specify this better
        """
        return self.dir_path.parent / self.imgtools_dir / file_name

    def _crawl(self) -> dict:
        # CRAWLER
        # -------
        # Checks if dataset has already been indexed
        if not self.csv_path.exists() or self.update:
            logger.debug("Output exists, force updating.", path=self.csv_path)
            logger.info("Indexing the dataset...")
            db = crawl(
                top=self.dir_path,
                n_jobs=self.n_jobs,
                csv_path=self.csv_path,
                json_path=self.json_path,
            )
            logger.info(f"Number of patients in the dataset: {len(db)}")
        else:
            logger.warning(
                "The dataset has already been indexed. Use --update to force update."
            )
        return db

    def parse_graph(self, modalities: str | List[str]) -> pd.DataFrame:
        """Parse the graph based on the provided modalities.

        TODO: establish how users can use the df columns to filter the data.

        Parameters
        ----------
        modalities : str | List[str]
            Either a list of strings or a single string of comma-separated modalities.
            For example, ["CT", "RTSTRUCT"] or "CT,RTSTRUCT".
            Only samples with ALL specified modalities will be processed.
            Ensure there are no spaces between list elements if provided as a string.

        Returns
        -------
        pd.DataFrame
            A pandas DataFrame containing the parsed graph.
        """
        assert self.graph is not None, "Graph not initialized."
        modalities = (
            modalities if isinstance(modalities, str) else ",".join(modalities)
        )
        logger.info("Querying graph", modality=modalities)
        self.df_combined = self.graph.parser(modalities)
        return self.df_combined

    def _init_graph(self) -> DataGraph:
        # GRAPH
        # -----
        # Form the graph
        logger.debug("Creating edge path", edge_path=self.edge_path)
        return DataGraph(
            path_crawl=self.csv_path.resolve().as_posix(),
            edge_path=self.edge_path.as_posix(),
            update=self.update_crawl,
        )

    def __call__(self, key: object) -> object:
        """Retrieve input data."""
        return self._loader.get(key)


# TODO: these two are useful, but need some work.
#       Figure out how to best make them useful for the user.


class ImageCSVInput(BaseInput):
    """
    ImageCSVInput class for loading images from a CSV file.

    Parameters
    ----------
    csv_path_or_dataframe : str
        Path to the CSV file or a pandas DataFrame.

    colnames : List[str]
        Column names in the CSV file for image loading.

    id_column : Optional[str]
        Column name to use as the subject ID. Default is None.

    expand_paths : bool
        Whether to expand relative paths. Default is True.

    readers : List[Callable]
        Functions to read images. Default is [read_image].
    """

    def __init__(
        self,
        csv_path_or_dataframe: str,
        colnames: List[str],
        id_column: Optional[str] = None,
        expand_paths: bool = True,
        readers: Optional[List] = None,
    ) -> None:
        self.csv_path_or_dataframe = csv_path_or_dataframe
        self.colnames = colnames
        self.id_column = id_column
        self.expand_paths = expand_paths
        self.readers = readers or [read_image]
        self._loader = ImageCSVLoader(
            self.csv_path_or_dataframe,
            colnames=self.colnames,
            id_column=self.id_column,
            expand_paths=self.expand_paths,
            readers=self.readers,
        )

    def __call__(self, key: Any) -> Any:  # noqa: ANN401
        """Retrieve input data."""
        return self._loader.get(key)


class ImageFileInput(BaseInput):
    """
    ImageFileInput class for loading images from a file directory.

    Parameters
    ----------
    root_directory : str
        Root directory where the image files are stored.

    get_subject_id_from : str
        Method for extracting subject IDs. Options are 'filename' or 'subject_directory'.

    subdir_path : Optional[str]
        Subdirectory path for images. Default is None.

    exclude_paths : List[str]
        List of paths to exclude. Default is an empty list.

    reader : Callable
        Function to read images. Default is `read_image`.
    """

    def __init__(
        self,
        root_directory: str,
        get_subject_id_from: str = "filename",
        subdir_path: Optional[str] = None,
        exclude_paths: Optional[List[str]] = None,
        reader: Optional[Any] = None,  # noqa: ANN401
    ) -> None:
        self.root_directory = root_directory
        self.get_subject_id_from = get_subject_id_from
        self.subdir_path = subdir_path
        self.exclude_paths = exclude_paths or []
        self.reader = reader or read_image
        self._loader = ImageFileLoader(
            self.root_directory,
            get_subject_id_from=self.get_subject_id_from,
            subdir_path=self.subdir_path,
            exclude_paths=self.exclude_paths,
            reader=self.reader,
        )

    def __call__(self, key: Any) -> Any:  # noqa: ANN401
        """Retrieve input data."""
        return self._loader.get(key)


# ruff: noqa
if __name__ == "__main__":  # pragma: no cover
    # Demonstrating usage with mock data
    example_csv_input = ImageCSVInput(
        "path/to/csv", colnames=["image", "label"]
    )
    print("ExampleCSVInput repr:", repr(example_csv_input))

    example_file_input = ImageFileInput("path/to/directory")
    print("ExampleFileInput repr:", repr(example_file_input))



from .utils import (
    DicomInput,
    extract_roi_meta,
    extract_roi_names,
    load_rtstruct_dcm,
    rtstruct_reference_uids,
)

__all__ = [
    # utils
    "DicomInput",
    "extract_roi_meta",
    "extract_roi_names",
    "load_rtstruct_dcm",
    "rtstruct_reference_uids",
]



from unittest.mock import MagicMock, patch

import numpy as np
import pytest
import SimpleITK as sitk

from imgtools.modules.structureset import (
    StructureSet,
)  # Replace `your_module` with the actual module name


@pytest.fixture
def roi_points():
    """Fixture for mock ROI points."""
    return {
        "GTV": [np.array([[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]])],
        "PTV": [np.array([[2.0, 2.0, 2.0], [3.0, 3.0, 3.0]])],
        "CTV_0": [np.array([[4.0, 4.0, 4.0], [5.0, 5.0, 5.0]])],
        "CTV_1": [np.array([[6.0, 6.0, 6.0], [7.0, 7.0, 7.0]])],
        "CTV_2": [np.array([[8.0, 8.0, 8.0], [9.0, 9.0, 9.0]])],
        "ExtraROI": [np.array([[10.0, 10.0, 10.0], [11.0, 11.0, 11.0]])],
    }


@pytest.fixture
def metadata():
    """Fixture for mock metadata."""
    return {"PatientName": "John Doe"}


# Parametrized tests for simple and moderately complex cases
@pytest.mark.parametrize(
    "names, roi_select_first, roi_separate, expected",
    [
        # Case 1: Default behavior with exact matches
        (["GTV", "PTV"], False, False, {"GTV": 0, "PTV": 1}),
        # Case 2: Regex matching
        (["GTV", "P.*"], False, False, {"GTV": 0, "PTV": 1}),
        # Case 3: Select only the first match for each pattern
        (["G.*", "P.*"], True, False, {"GTV": 0, "PTV": 1}),
        # Case 4: Separate matches for regex pattern
        (["P.*"], False, True, {"PTV": 0}),
        # Case 5: Regex pattern with multiple matches (consolidated labels)
        (["CTV.*"], False, False, {"CTV_0": 0, "CTV_1": 0, "CTV_2": 0}),
        # Case 6: Regex pattern with multiple matches (separate labels)
        (["CTV.*"], False, True, {"CTV_0": 0, "CTV_1": 0, "CTV_2": 0}),
        # Case 7: Grouped patterns
        (
            [["GTV", "PTV"], "CTV.*"],
            False,
            False,
            {"GTV": 0, "PTV": 0, "CTV_0": 1, "CTV_1": 1, "CTV_2": 1},
        ),
        # Case 8: Grouped patterns with separate labels for regex matches
        # ([["GTV", "PTV"], "CTV.*"], False, True, {"GTV": 0, "PTV": 0, "CTV_0": 1, "CTV_1": 2, "CTV_2": 3}),
    ],
)
def test_assign_labels(
    names, roi_select_first, roi_separate, expected, roi_points
) -> None:
    """Test _assign_labels method with various cases."""
    structure_set = StructureSet(roi_points)
    result = structure_set._assign_labels(
        names, roi_select_first, roi_separate
    )
    assert result == expected


# Parametrized tests for complex scenarios with intricate patterns
@pytest.mark.parametrize(
    "names, roi_select_first, roi_separate, expected",
    [
        # Case 1: Complex regex patterns with partial matches
        (
            ["G.*", "C.*1", "Extra.*"],
            False,
            False,
            {"GTV": 0, "CTV_1": 1, "ExtraROI": 2},
        ),
        # Case 2: Nested regex patterns with grouped and separated labels
        (
            [["GTV", "CTV.*"], "P.*", "Extra.*"],
            False,
            False,
            {
                "GTV": 0,
                "CTV_0": 0,
                "CTV_1": 0,
                "CTV_2": 0,
                "PTV": 1,
                "ExtraROI": 2,
            },
        ),
        # ([["GTV", "CTV.*"], "P.*", "Extra.*"], False, True, {"GTV": 0, "CTV_0_0": 1, "CTV_1_1": 2, "CTV_2_2": 3, "PTV": 4, "ExtraROI": 5}),
        # Case 3: Regex patterns that match all ROIs
        (
            [".*"],
            False,
            False,
            {
                "GTV": 0,
                "PTV": 0,
                "CTV_0": 0,
                "CTV_1": 0,
                "CTV_2": 0,
                "ExtraROI": 0,
            },
        ),
        # ([".*"], False, True, {"GTV_0": 0, "PTV_1": 1, "CTV_0_2": 2, "CTV_1_3": 3, "CTV_2_4": 4, "ExtraROI_5": 5}),
        # Case 4: Overlapping regex patterns
        (
            ["G.*", "C.*", "Extra.*"],
            False,
            False,
            {"GTV": 0, "CTV_0": 1, "CTV_1": 1, "CTV_2": 1, "ExtraROI": 2},
        ),
        # (["G.*", "C.*", "Extra.*"], False, True, {"GTV": 0, "CTV_0_0": 1, "CTV_1_1": 2, "CTV_2_2": 3, "ExtraROI_3": 4}),
        # Case 5: No matches for given patterns
        pytest.param(
            ["NonExistent.*"],
            False,
            False,
            {},
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        # Case 6: Conflicting options (should raise an error)
        # pytest.param(["G.*"], True, True, None, marks=pytest.mark.xfail(raises=ValueError)),
    ],
)
def test_assign_labels_complex(
    names, roi_select_first, roi_separate, expected, roi_points
) -> None:
    """Test _assign_labels method with complex scenarios."""
    structure_set = StructureSet(roi_points)
    result = structure_set._assign_labels(
        names, roi_select_first, roi_separate
    )
    assert result == expected


def test_assign_labels_invalid(roi_points) -> None:
    """Test _assign_labels method with invalid inputs."""
    structure_set = StructureSet(roi_points)

    # Case: Empty names
    with pytest.raises(ValueError, match="The 'names' list cannot be empty."):
        structure_set._assign_labels([])

    # Case: Conflicting options
    with pytest.raises(
        ValueError,
        match="The options 'roi_select_first' and 'roi_separate' cannot both be True.",
    ):
        structure_set._assign_labels(
            ["G.*"], roi_select_first=True, roi_separate=True
        )


def test_init(roi_points, metadata) -> None:
    """Test StructureSet initialization."""
    structure_set = StructureSet(roi_points, metadata)
    assert structure_set.roi_points == roi_points
    assert structure_set.metadata == metadata

    # Test default metadata
    structure_set_no_metadata = StructureSet(roi_points)
    assert structure_set_no_metadata.metadata == {}


@patch("imgtools.modules.structureset.dcmread")
def test_from_dicom_rtstruct(mock_dcmread) -> None:
    """Test from_dicom_rtstruct method with mocked DICOM file."""
    """Test from_dicom_rtstruct method with mocked DICOM file."""
    mock_rtstruct = MagicMock()
    mock_rtstruct.StructureSetROISequence = [
        MagicMock(ROIName="GTV"),
        MagicMock(ROIName="PTV"),
    ]
    mock_rtstruct.ROIContourSequence = [
        MagicMock(),
        MagicMock(),
    ]
    mock_rtstruct.ROIContourSequence[0].ContourSequence = [
        MagicMock(ContourData=[1.0, 2.0, 3.0])
    ]
    mock_rtstruct.ROIContourSequence[1].ContourSequence = [
        MagicMock(ContourData=[4.0, 5.0, 6.0])
    ]
    mock_rtstruct.Modality = "RTSTRUCT"
    mock_dcmread.return_value = mock_rtstruct

    structure_set = StructureSet.from_dicom_rtstruct("dummy")
    # Assert the results
    assert "GTV" in structure_set.roi_points
    assert "PTV" in structure_set.roi_points
    assert len(structure_set.roi_points["GTV"]) == 1
    assert len(structure_set.roi_points["PTV"]) == 1


def test_rtstruct_to_segmentation(roi_points, metadata) -> None:
    """Test rtstruct_to_segmentation method with mocked ROI points."""
    structure_set = StructureSet(roi_points, metadata)
    # reference sitk_image
    ref_image = sitk.Image(10, 10, 10, sitk.sitkFloat32)

    seg_images = structure_set.to_segmentation(
        reference_image=ref_image, roi_names=["GTV", "PTV"]
    )

    assert seg_images is not None

    raw_roi_names = seg_images.raw_roi_names

    assert "GTV" in raw_roi_names
    assert "PTV" in raw_roi_names
    assert list(seg_images.roi_indices.keys()) == ["GTV", "PTV"]

    seg_images2 = structure_set.to_segmentation(
        reference_image=ref_image, roi_names="GTV"
    )

    assert seg_images2 is not None

    assert "GTV" in seg_images2.raw_roi_names
    assert "PTV" not in seg_images2.raw_roi_names

    assert seg_images2.get_label(name="GTV") is not None

    assert list(seg_images2.roi_indices.keys()) == ["GTV"]

    assert repr(seg_images2) == "<Segmentation with ROIs: {'GTV': 1}>"
